# -*- coding: utf-8 -*-
import LeProHQ
import numpy as np
from scipy.integrate import quad

from ..partonic_channel import RSL
from . import partonic_channel as pc


class GluonVV(pc.NeutralCurrentBase):
    def NLO(self):
        """
        |ref| implements :eqref:`D.1`, :cite:`felix-thesis`.
        """

        def cg(z, _args):
            if self.is_below_threshold(z):
                return 0.0
            return (
                self._FHprefactor / z * LeProHQ.cg0("F2", "VV", self._xi, self._eta(z))
            )

        return RSL(cg)

    def NNLO(self):
        """
        |ref| implements NLO (heavy) gluon coefficient function, :cite:`felix-thesis`.
        """

        def cg(z, _args):
            if self.is_below_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cg1("F2", "VV", self._xi, self._eta(z))
                    + LeProHQ.cgBar1("F2", "VV", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cg)


class GluonAA(GluonVV):
    def NLO(self):
        """
        |ref| implements :eqref:`D.4`, :cite:`felix-thesis`.
        """

        def cg(z, _args):
            if self.is_below_threshold(z):
                return 0.0
            return (
                self._FHprefactor / z * LeProHQ.cg0("F2", "AA", self._xi, self._eta(z))
            )

        return RSL(cg)

    def NNLO(self):
        """
        |ref| implements NLO (heavy) gluon coefficient function, :cite:`felix-thesis`.
        """

        def cg(z, _args):
            if self.is_below_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cg1("F2", "AA", self._xi, self._eta(z))
                    + LeProHQ.cgBar1("F2", "AA", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cg)


class SingletVV(pc.NeutralCurrentBase):
    def NNLO(self):
        """
        |ref| implements NLO (heavy) singlet coefficient function, :cite:`felix-thesis`.
        """

        def cq(z, _args):
            if self.is_below_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cq1("F2", "VV", self._xi, self._eta(z))
                    + LeProHQ.cqBarF1("F2", "VV", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cq)


class SingletAA(pc.NeutralCurrentBase):
    def NNLO(self):
        """
        |ref| implements NLO (heavy) singlet coefficient function, :cite:`felix-thesis`.
        """

        def cq(z, _args):
            if self.is_below_threshold(z):
                return 0.0
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * (
                    LeProHQ.cq1("F2", "AA", self._xi, self._eta(z))
                    + LeProHQ.cqBarF1("F2", "AA", self._xi, self._eta(z))
                    * np.log(self._xi)
                )
            )

        return RSL(cq)


class NonSinglet(pc.NeutralCurrentBase):
    def NNLO(self):
        """
        |ref| implements NLO (heavy) non-singlet coefficient function, :cite:`felix-thesis`.
        """

        def dq(z, _args):
            if self.is_below_threshold(z):
                return 0.0
            # TODO move this hack into LeProHQ
            eta = self._eta(z)
            eta = min(eta, 1e5)
            return (
                self._FHprefactor
                / z
                * (4.0 * np.pi) ** 2
                * LeProHQ.dq1("F2", "VV", self._xi, eta)
            )

        def Adler(_x, _args):
            l = quad(dq, 0.0, 1.0, args=np.array([]))
            return -l[0]

        return RSL(dq, loc=Adler)



# -*- coding: utf-8 -*-
import numpy as np

from ..observable_name import ObservableName
from .esf import ESFInfo
from .result import ESFResult, EXSResult

GEV_CM2_CONV = 3.893793e10
"Conversion factor from GeV^-2 to 10^-38 cm^2"


def xs_coeffs(kind, y, x=None, Q2=None, params=None):
    """Compute coefficients in the definition of a given cross section.

    Parameters
    ----------
    kind : str
        the identifier of the cross section for which coefficients have to be
        computed
    y : float
        Bjorken y value for the chosen kinematic point
    x: None or float
        Bjorken x value for the chosen kinematic point
    Q2: None or float
        photon virtuality for the chosen kinematic point
    params : None or dict
        theory parameters required to fully specify the coefficients (which ones
        depends on the chosen definition, i.e. on `kind`)

    Returns
    -------
    np.ndarray
        the coefficients of the cross-section on the basis `(F2, FL, xF3)`

    Raises
    ------
    ValueError or KeyError
        in case not enough parameters have been specified

    """
    yp = 1.0 + (1.0 - y) ** 2
    ym = 1.0 - (1.0 - y) ** 2
    yL = y**2

    # Neutral Currents
    # HERA average
    if kind == "XSHERANCAVG":
        return np.array([1.0, -yL / yp, 0.0])

    if params is None:
        raise ValueError(f"Parameters required to compute '{kind}' cross section.")

    f3sign = -1 if params["projectilePID"] < 0 else 1
    # HERA fixed lepton
    if kind == "XSHERANC":
        return np.array([1.0, -yL / yp, f3sign * ym / yp])

    # Charged Currents
    norm = 0.0
    # HERA
    if kind == "XSHERACC":
        norm = 1.0 / 4.0
    # FW CDHSW see http://cds.cern.ch/record/200223/files/198909396.pdf
    elif kind == "FW":
        mn = np.sqrt(params["M2target"])
        norm = 1.0
        ym = 0
        yp = 1.0
        yL = y**2 / (2 * (y**2 / 2 + (1 - y) - (mn * x * y) ** 2 / Q2))
    else:
        mn = np.sqrt(params["M2target"])
        m2w = params["M2W"]
        yp -= 2.0 * (mn * x * y) ** 2 / Q2  # = ypc
        # CHORUS
        if kind == "XSCHORUSCC":
            norm = (
                GEV_CM2_CONV
                * params["GF"] ** 2
                * mn
                / (2.0 * np.pi * (1.0 + Q2 / m2w) ** 2)
            )
        # NUTEV
        if kind == "XSNUTEVCC":
            norm = 100.0 / 2.0 / (1.0 + Q2 / m2w) ** 2
        # NUTEV neutrino dis
        if kind == "XSNUTEVNU":
            norm = (
                GEV_CM2_CONV
                * params["GF"] ** 2
                * mn
                / (2.0 * np.pi)
            )
    return np.array([yp, -yL, f3sign * ym]) * norm


class EvaluatedCrossSection:
    def __init__(self, kin, obs_name, configs, get_esf):
        self.kin = kin
        self.Q2 = kin["Q2"]
        self.x = kin["x"]
        self.y = kin["y"]
        self.info = ESFInfo(obs_name, configs)
        self.get_esf = get_esf

    def alpha_qed_power(self):
        return 0

    def get_result(self):
        # Collect esfs
        flavor = self.info.obs_name.flavor
        f_coeffs = xs_coeffs(
            self.info.obs_name.kind,
            self.y,
            x=self.x,
            Q2=self.Q2,
            params=dict(
                projectilePID=self.info.configs.coupling_constants.obs_config[
                    "projectilePID"
                ],
                M2target=self.info.configs.M2target,
                M2W=self.info.configs.M2W,
                GF=self.info.configs.GF,
            ),
        )
        f2 = self.get_esf(ObservableName(f"F2_{flavor}"), self.kin).get_result()
        fl = self.get_esf(ObservableName(f"FL_{flavor}"), self.kin).get_result()
        # skip F3 if it is not required
        if f_coeffs[2] != 0.0:
            f3 = self.get_esf(ObservableName(f"F3_{flavor}"), self.kin).get_result()
        else:
            f3 = ESFResult(self.x, self.Q2, f2.nf)

        # add normalizations
        esf = f_coeffs @ np.array([f2, fl, f3])
        # remap to EXS
        sigma = EXSResult(self.x, self.Q2, self.y, f2.nf)
        for o, v in esf.orders.items():
            # now shift orders: push alpha_qed two powers up
            sigma.orders[(o[0], o[1] + self.alpha_qed_power(), o[2], o[3])] = v
        return sigma



# -*- coding: utf-8 -*-
import numpy as np
from banana.benchmark.external import apfel_utils


def load_apfel(theory, observables, pdf="ToyLH", use_external_grid=False):
    """
    Set APFEL parameter from ``theory`` dictionary.

    Parameters
    ----------
        theory : dict
            theory and process parameters
        observables : dict
            observables runcard
        pdf : str
            PDF name

    Returns
    -------
        module
            loaded apfel wrapper
    """
    apfel = apfel_utils.load_apfel(
        theory, observables, pdf, use_external_grid=use_external_grid
    )

    # set DIS params
    apfel.SetProcessDIS(observables.get("prDIS", "EM"))
    apfel.SetPropagatorCorrection(observables.get("PropagatorCorrection", 0))
    apfel.SetPolarizationDIS(observables.get("PolarizationDIS", 0))
    apfel.SetProjectileDIS(observables.get("ProjectileDIS", "electron"))
    apfel.SetTargetDIS(observables.get("TargetDIS", "proton"))
    charge = observables.get("NCPositivityCharge")
    if charge is not None:
        apfel.SelectCharge(charge)

    # apfel initialization for DIS
    apfel.InitializeAPFEL_DIS()

    return apfel


def compute_apfel_data(theory, observables, pdf):
    """
    Run APFEL to compute observables.

    Parameters
    ----------
        theory : dict
            theory runcard
        observables : dict
            observables runcard
        pdf : lahapdf_like
            PDF set

    Returns
    -------
        apf_tab : dict
            AFPEL numbers
    """
    # setup APFEL
    apfel = load_apfel(theory, observables, pdf.set().name)

    # mapping observables names to APFEL methods
    apfel_structure_functions = {
        "F2_light": apfel.F2light,
        "FL_light": apfel.FLlight,
        "F3_light": apfel.F3light,
        "F2_charm": apfel.F2charm,
        "F2_bottom": apfel.F2bottom,
        "F2_top": apfel.F2top,
        "FL_charm": apfel.FLcharm,
        "FL_bottom": apfel.FLbottom,
        "FL_top": apfel.FLtop,
        "F3_charm": apfel.F3charm,
        "F3_bottom": apfel.F3bottom,
        "F3_top": apfel.F3top,
        "F2_total": apfel.F2total,
        "FL_total": apfel.FLtotal,
        "F3_total": apfel.F3total,
        "F2": apfel.F2total,
        "FL": apfel.FLtotal,
        "F3": apfel.F3total,
    }

    l = ""
    if observables["ProjectileDIS"] == "electron":
        l = "E"
    elif observables["ProjectileDIS"] == "positron":
        l = "P"
    elif observables["ProjectileDIS"] == "neutrino":
        l = "NU"
    elif observables["ProjectileDIS"] == "antineutrino":
        l = "NB"

    apfel_fkobservables = {
        "XSHERANC_light": f"DIS_NC{l}_L",
        "XSHERANC_charm": f"DIS_NC{l}_CH",
        "XSHERANC_bottom": f"DIS_NC{l}_BT",
        "XSHERANC_top": f"DIS_NC{l}_TP",
        "XSHERANC_total": f"DIS_NC{l}",
        "XSHERANC": f"DIS_NC{l}",
        "XSHERANCAVG_charm": f"DIS_NC{l}_CH",
        "XSHERANCAVG_bottom": f"DIS_NC{l}_BT",
        "XSHERANCAVG_top": f"DIS_NC{l}_TP",
        "XSHERACC_light": f"DIS_CC{l}_L",
        "XSHERACC_charm": f"DIS_CC{l}_CH",
        "XSHERACC_bottom": f"DIS_CC{l}_BT",
        "XSHERACC_top": f"DIS_CC{l}_TP",
        "XSHERACC_total": f"DIS_CC{l}",
        "XSHERACC": f"DIS_CC{l}",
        "XSCHORUSCC_light": f"DIS_S{l}_L",
        "XSCHORUSCC_charm": f"DIS_S{l}_C",
        "XSCHORUSCC_total": f"DIS_S{l}",
        "XSCHORUSCC": f"DIS_S{l}",
        "XSNUTEVCC_charm": f"DIS_DM_{l}",
    }

    # compute observables with APFEL
    apf_tab = {}
    for obs_name, kinematics in observables["observables"].items():
        apf_tab[obs_name] = []
        # a cross section?
        if obs_name in apfel_fkobservables:
            apf_obs = apfel_fkobservables[obs_name]
            if observables["TargetDIS"] == "lead":
                apf_obs += "_Pb"
            # FK calls SetProcessDIS, SetProjectileDIS and SetTargetDIS
            apfel.SetFKObservable(apf_obs)
        elif obs_name not in apfel_structure_functions:  # not a SF?
            raise ValueError(f"Unknown observable {obs_name}")

        # iterate over input kinematics
        for kin in kinematics:
            Q2 = kin["Q2"]
            x = kin["x"]
            y = kin["y"]

            # disable APFEL evolution: we are interested in the pure DIS part
            #
            # setting initial scale to muF (sqrt(Q2)*xiF) APFEL is going to:
            # - take the PDF at the scale of muF (exactly as we are doing)
            # - evolve from muF to muF because the final scale is the second
            #   argument times xiF (internally), so actually it's not evolving
            apfel.ComputeStructureFunctionsAPFEL(
                np.sqrt(Q2) * theory["XIF"], np.sqrt(Q2)
            )
            # compute the actual result
            if obs_name in apfel_structure_functions:
                result = apfel_structure_functions[obs_name](x)
            else:
                result = apfel.FKObservables(x, np.sqrt(Q2), y)
            # take over the kinematics
            r = kin.copy()
            r["result"] = result
            apf_tab[obs_name].append(r)

    return apf_tab



# -*- coding: utf-8 -*-
"""
The coefficient functions definition is given in :eqref:`4.2`,
:cite:`vogt-f2nc` (that is the main reference for their expression, i.e. all
the formulas in this module) (the same of :eqref:`1` in :cite:`vogt-flnc`).
The main reference for their expression is :cite:`vogt-flnc`.

Scale varitions main reference is :cite:`nnlo-sv-singlet` and
:cite:`nnlo-sv-nonsinglet`.

Scale variations main reference is :cite:`moch-f3nc`.

Note
----
    Check the theory reference for details on
    :doc:`../theory/scale-variations`
"""

from . import kernels



# -*- coding: utf-8 -*-
"""
Provides the Altarelli-Parisi splitting functions.

The splitting functions are defined by :cite:`split-ns,split-singlet` and
they are organized by the outgoing and incoming particle.
Furthermore they are organized according to their RSL structure.

"""

import numba as nb
import numpy as np
from eko import constants

from ..partonic_channel import RSL


@nb.njit("f8(f8,f8[:])", cache=True)
def pqq_reg(z, _args):
    """
    Regular part of :math:`P_{qq}` splitting function.

    |ref| implements :eqref:`4.5`, :cite:`split-ns`.

    Parameters
    ----------
        z : float
            partonic momentum fraction

    Returns
    -------
        float
            the regular bit of quark-quark splitting function :math:`P_{qq}^R(z)`

    """
    return -2.0 * constants.CF * (1.0 + z)


@nb.njit("f8(f8,f8[:])", cache=True)
def pqq_sing(z, _args):
    """
    Singular part of :math:`P_{qq}` splitting function.

    |ref| implements :eqref:`4.5`, :cite:`split-ns`.

    Parameters
    ----------
        z : float
            partonic momentum fraction

    Returns
    -------
        float
            the singular bit the quark-quark splitting function :math:`P_{qq}^S(z)`
    """
    return 4.0 * constants.CF / (1.0 - z)


@nb.njit("f8(f8,f8[:])", cache=True)
def pqq_local(x, _args):
    r"""
    Local part of :math:`P_{qq}` splitting function.

    |ref| implements :eqref:`4.5`, :cite:`split-ns`.

    Parameters
    ----------
        x : float
            hadronic momentum fraction

    Returns
    -------
        float
            the local bit of quark-quark splitting function :math:`P_{qq}^L(x)`

    """
    return constants.CF * (3.0 + 4.0 * np.log(1.0 - x))


def pqq(_nf):
    """
    :math:`P_{qq}` splitting function.

    |ref| implements :eqref:`4.5`, :cite:`split-ns`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        RSL
            the quark-quark splitting function :math:`P_{qq}(z)`
    """
    return RSL(pqq_reg, pqq_sing, pqq_local)


@nb.njit("f8(f8,f8[:])", cache=True)
def pqg_single(z, _args):
    """
    :math:`P_{q_ig}` splitting function (i.e.: :math:`P_{qg}` for a single
    flavor).

    |ref| implements :eqref:`4.6` and :eqref:`2.5`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction

    Returns
    -------
        float
            the quark-gluon splitting function :math:`P_{q_ig}(z)`

    """
    return 2.0 * constants.TR * (z**2 + (1.0 - z) ** 2)


@nb.njit("f8(f8,f8[:])", cache=True)
def pqg_reg(z, args):
    """
    (Regular) :math:`P_{qg}` splitting function.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction

    Returns
    -------
        float
            the quark-gluon splitting function :math:`P_{qg}(z)`

    """
    return 2 * args[0] * pqg_single(z, args)


def pqg(nf):
    """
    :math:`P_{qg}` splitting function.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        nf : int
            number of active flavors

    Returns
    -------
        RSL
            the quark-gluon splitting function :math:`P_{qg}(z)`

    """
    return RSL(pqg_reg, args=[nf])


raw_labels = {"P_qq_0": pqq, "P_qg_0": pqg}



# -*- coding: utf-8 -*-
"""
Goal
----

These subpackage has the main goal of parsing and validate user input, in
particular the following three operations are performed:

#. **constraints**: constraints on single fields are enforced, raising an error
    for each violation
#. **cross-constraints**: constraints involving multiple fields are enforced,
    raising an error for each violation

Note
----

We currently chose not to use any default, in order to force the user to provide a
fully explicit input.
Even if in this way it will be a little bit more difficult to learn how to use
`yadism` it is not intended to be used by people not aware of its inputs.

Instead we are going to provide one or more example inputs in the project (or
project docs) and each field is fully described inside this subpackage in
:mod:`domains.yaml`.

"""



# -*- coding: utf-8 -*-
import numpy as np

from . import fonll, heavy, intrinsic, kernels, light
from .partonic_channel import EmptyPartonicChannel


class Component(list):
    """
    Used for organize elements and debugging purpose.
    """

    heavyness = {0: "light", 4: "charm", 5: "bottom", 6: "top"}

    def __init__(self, heavy, kernels=None):
        self.heavy = self.heavyness[heavy]

        if kernels is None:
            kernels = []
        super().__init__(kernels)

    def __repr__(self):
        return self.heavy + f"({len(self)} kernels)"


# TODO add more doc strings


class Combiner:
    """
    Does the matching between coefficient functions and partons with their approptiate coupling
    strength.

    Parameters
    ----------
        esf : EvaluateStructureFunction
            current ESF
    """

    def __init__(self, esf):
        self.esf = esf
        self.masses = {4 + i: not mass for i, mass in enumerate(esf.info.ZMq)}
        self.intrinsic = esf.info.intrinsic_range
        self.obs_name = esf.info.obs_name
        self.nf = esf.info.threshold.nf(esf.Q2)
        self.target = esf.info.target

    def collect(self):
        comps = []

        family = self.obs_name.flavor_family

        # Adding light component
        if family in ["light", "total"]:
            comps.append(self.light_component())
        if family == "heavy":
            #  the only case in which an heavy contribution is not present in those
            #  accounted for in total, it's whene heavy already became heavylight
            comps.extend(self.heavylight_components())
        if family in ["heavy", "total"]:
            comps.extend(self.heavy_components())

        return comps

    def light_component(self):
        nf = self.nf
        masses = self.masses

        comp = Component(0)

        # the first condition essentially checks nf != 3
        if nf in masses and masses[nf]:
            nl = nf - 1
            comp.extend(
                fonll.kernels.generate_light(
                    self.esf, nl, self.esf.info.theory["pto_evol"]
                )
            )
            comp.extend(heavy.kernels.generate_missing(self.esf, nl, nl + 1))
            comp.extend(
                self.damp_elems(
                    nl,
                    fonll.kernels.generate_light_diff(
                        self.esf,
                        nl,
                        self.esf.info.theory["pto_evol"],
                    ),
                )
            )
        else:
            comp.extend(light.kernels.generate(self.esf, nf))

        for ihq in range(nf + 1, 7):
            if masses[ihq]:
                comp.extend(heavy.kernels.generate_missing(self.esf, nf, ihq))
        return comp

    def heavylight_components(self):
        nf = self.nf
        hq = self.obs_name.hqnumber
        masses = self.masses

        comps = []

        if hq < nf or (hq == nf and not masses[hq]):
            heavylight = Component(hq)
            heavylight.extend(kernels.generate_single_flavor_light(self.esf, nf, hq))
            comps.append(heavylight)

        return comps

    def heavy_components(self):
        nf = self.nf
        hq = self.obs_name.hqnumber
        masses = self.masses

        comps = []

        heavy_comps = {}
        for sfh in range(nf, 7):
            # exclude sfh=3, since heavy contributions are there for [4,5,6]
            # if it's ZM you don't even have the component
            if sfh not in masses or not masses[sfh]:
                continue

            heavy_comps[sfh] = Component(sfh)
            if hq not in (0, sfh):
                continue

            if sfh == nf:
                # then it is FONLL
                nl = nf - 1
                heavy_comps[sfh].extend(heavy.kernels.generate(self.esf, nl, ihq=sfh))
                if sfh not in self.intrinsic:
                    heavy_comps[sfh].extend(
                        self.damp_elems(
                            nl,
                            fonll.kernels.generate_heavy_diff(
                                self.esf, nl, self.esf.info.theory["pto_evol"]
                            ),
                        )
                    )
                else:
                    heavy_comps[sfh].extend(
                        intrinsic.kernels.generate(self.esf, ihq=sfh)
                    )
                    heavy_comps[sfh].extend(
                        self.damp_elems(
                            nl,
                            fonll.kernels.generate_heavy_intrinsic_diff(
                                self.esf, nl, self.esf.info.theory["pto_evol"]
                            ),
                        )
                    )
            else:
                # then it is *not* FONLL
                heavy_comps[sfh].extend(heavy.kernels.generate(self.esf, nf, ihq=sfh))
                if sfh in self.intrinsic:
                    heavy_comps[sfh].extend(
                        intrinsic.kernels.generate(self.esf, ihq=sfh)
                    )

            for ihq in range(sfh + 1, 7):
                if masses[ihq]:
                    heavy_comps[sfh].extend(
                        heavy.kernels.generate_missing(self.esf, nf, ihq, icoupl=sfh)
                    )
            comps.append(heavy_comps[sfh])

        return comps

    def damp_elems(self, nl, elems):
        """
        Damp FONLL difference contributions if necessary.

        Parameters
        ----------
            nl : int
                number of *light* flavors
            elems : list(Kernel)
                kernels to be modified

        Returns
        -------
            elems : list(Kernel)
                modified kernels
        """
        if not self.esf.info.FONLL_damping:
            return elems
        nhq = nl + 1
        # TODO: replace mass with threshold?
        m2hq = self.esf.info.m2hq[nhq - 4]
        power = self.esf.info.damping_powers[nhq - 4]
        if self.esf.Q2 > m2hq:
            damp = np.power(1.0 - m2hq / self.esf.Q2, power)
        else:
            damp = 0.0
        return (damp * e for e in elems)

    @staticmethod
    def apply_isospin(full, z, a):
        """
        Apply isospin symmetry to u and d distributions.

        Parameters
        ----------
            full : list(yadism.kernels.Kernel)
                all participants
            z : float
                number of protons
            a : float
                atomic mass number
        """
        nucl_factors = np.array([[z, a - z], [a - z, z]]) / a
        for ker in full:
            for sign in [-1, 1]:
                ps = np.array(
                    [ker.partons.get(sign * 1, 0), ker.partons.get(sign * 2, 0)]
                )
                ker.partons[sign * 1], ker.partons[sign * 2] = nucl_factors @ ps

    @staticmethod
    def drop_empty(full):
        """
        Drop kernels with :class:`EmptyPartonicChannel` or its partons with empty weight.

        Parameters
        ----------
            elems : list(yadism.kernels.Kernel)
                all participants

        Returns
        -------
            filtered_kernels : list(yadism.kernels.Kernel)
                improved participants
        """
        filtered_kernels = []
        for ker in full:
            ker.partons = {p: w for p, w in ker.partons.items() if w != 0}
            if len(ker.partons) > 0 and not isinstance(ker.coeff, EmptyPartonicChannel):
                filtered_kernels.append(ker)
        return filtered_kernels

    def collect_elems(self):
        """
        Collects all kernels according to the |FNS|.

        Returns
        -------
            elems : list(yadism.kernels.Kernel)
                all participants
        """
        components = self.collect()

        full = []
        for comp in components:
            full.extend(comp)

        # add level-0 nuclear correction: apply isospin symmetry
        self.apply_isospin(full, self.target["Z"], self.target["A"])

        # drop all kernels with 0 weight, or empty coeffs
        return self.drop_empty(full)



# -*- coding: utf-8 -*-
# pylint: skip-file
import abc

import numpy as np

from . import errors

# ┌───────────────────┐
# │ Arguments Classes │
# └───────────────────┘


class Argument(abc.ABC):
    """Base Class for input variables' domains semantics definition.

    Parameters
    ----------
    **kwargs
        Domain specification, with the syntax defined in ``domains.yaml``."""

    def __init__(self, **kwargs):
        """Load ``kwargs`` as object attributes."""
        for k, v in kwargs.items():
            setattr(self, k, v)

    @abc.abstractmethod
    def check_value(self, *, value):
        """Check if ``value`` belongs to the domain.

        Parameters
        ----------
        value : type
            Description of parameter `value`.
        """
        pass

    @abc.abstractmethod
    def _raise_error(self):
        """Raise the error, preprocessing the message."""
        pass


class StringArgument(Argument):
    """Argument definition for ``string`` type.

    ``string`` is the argument type for a generic string, and no other check is
    performed other than verifying that it is loaded just as a plain string and
    not any other kind, that should be managed by a more specific `Argument`
    checker.
    """

    def check_value(self, *, value):
        """Checks if `value` is actually a bare string.

        Raises
        ------
        DomainError
            If `value` is not in the domain of the ``type``.
        """

        if not isinstance(value, str):
            self._raise_error(value)

    def _raise_error(self, value):
        """Wrapper for DomainError, pretty print `domain`.

        No domain rule supported, if needed please choose a more specific
        argument type.
        """
        self.domain_ = ""
        raise errors.DomainError(value=value, **self.__dict__)


class EnumArgument(Argument):
    """Argument definition for ``enum`` type.

    ``enum`` type is just a domain defined by enumerating all the possible
    values.
    There is no hard-coded restrictions in having in the same enum domain values
    from different default types (e.g. integers and strings may be mixed).
    """

    def check_value(self, *, value):
        """Checks if `value` belongs to the domain.

        It checks if `value` is in the list of the enumerated values available
        for the ``type``.

        Raises
        ------
        DomainError
            If `value` is not in the domain of the ``type``.
        """

        if value not in self.domain:
            self._raise_error(value)

    def _raise_error(self, value):
        """Wrapper for DomainError, pretty print `domain`.

        The domain is formatted as a dashed list.
        """
        dom = list(map(str, self.domain))
        self.domain_ = "- " + "\n- ".join(dom)
        raise errors.DomainError(value=value, **self.__dict__)


class RealArgument(Argument):
    """Argument definition for ``real`` type.

    Actually just a real with restrictions on the domain, so an interval or the
    union of more disconnected intervals.
    """

    def check_value(self, *, value):
        """Checks if `value` belongs to the domain.

        It checks if `value` belongs to any interval available for the domain.

        Raises
        ------
        DomainError
            If `value` is not in the domain of the ``type``.
        """
        # load the value in the namespace with the name used in rules
        try:
            var_name = self.metavar
        except AttributeError:
            var_name = self.known_as

        # NOTE: the value is the only thing user provided, so it will be parsed
        # before to make sure it's nothing else than a float: with this check
        # `exec` is safe
        value = float(value)
        exec(f"{var_name} = {value}")

        # determine to which intervals value belongs to
        intervals = []
        for rule in self.domain:
            # NOTE: rule is a part of `yadism`, so `eval` is safe
            intervals += [eval(rule)]

        # NOTE: multiple conditions are in OR (not in AND)
        if not any(intervals):
            self._raise_error(value)

    def _raise_error(self, value):
        """Wrapper for DomainError, pretty print `domain`.

        The domain is formatted as a dashed list of rules.
        """
        dom = list(map(str, self.domain))
        self.domain_ = "- " + "\n- ".join(dom)
        raise errors.DomainError(value=value, **self.__dict__)


class IntegerArgument(RealArgument):
    """Short summary."""

    def check_value(self, *, value):
        value = float(value)
        value_int = np.round(value)
        if np.abs(value - value_int) > 1e-5:
            self._raise_error(value)
        super(IntegerArgument, self).check_value(value=value_int)


type_class_map = {
    "string": StringArgument,
    "enum": EnumArgument,
    "real": RealArgument,
    "integer": IntegerArgument,
}

# ┌───────────────────────────┐
# │ Cross Constraints Classes │
# └───────────────────────────┘


class CrossConstraint(abc.ABC):
    pass



# -*- coding: utf-8 -*-

import numba as nb

from ..intrinsic import fl_nc as intrinsic
from ..light import fl_nc as light
from ..partonic_channel import RSL, EmptyPartonicChannel
from . import partonic_channel as pc
from . import raw_nc


@nb.njit("f8(f8,f8[:])", cache=True)
def cg_NLL_NLO(z, _args):
    return raw_nc.clg1am0_a0(z)


@nb.njit("f8(f8,f8[:])", cache=True)
def cg_NLL_NNLO(z, args):
    L = args[0]
    return (raw_nc.clg2am0_aq(z) - raw_nc.clg2am0_af(z)) * L


@nb.njit("f8(f8,f8[:])", cache=True)
def cg_NNLL_NNLO(z, _args):
    return raw_nc.clg2am0_a0(z)


class AsyLLGluon(EmptyPartonicChannel):
    pass


class AsyNLLGluon(pc.PartonicChannelAsy):
    def NLO(self):
        return RSL(cg_NLL_NLO, args=[self.L])

    def NNLO(self):
        return RSL(cg_NLL_NNLO, args=[self.L])


class AsyNNLLGluon(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cg_NNLL_NNLO)


@nb.njit("f8(f8,f8[:])", cache=True)
def cps_NLL_NNLO(z, args):
    L = args[0]
    return (raw_nc.clps2am0_aq(z) - raw_nc.clps2am0_af(z)) * L


@nb.njit("f8(f8,f8[:])", cache=True)
def cps_NNLL_NNLO(z, _args):
    return raw_nc.clps2am0_a0(z)


class AsyLLSinglet(EmptyPartonicChannel):
    pass


class AsyNLLSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cps_NLL_NNLO, args=[self.L])


class AsyNNLLSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cps_NNLL_NNLO)


class AsyLLNonSinglet(EmptyPartonicChannel):
    pass


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_NLL_NNLO(z, args):
    L = args[0]
    return raw_nc.clns2am0_aq(z) * L


class AsyNLLNonSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cns_NLL_NNLO)


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_NNLL_NNLO(z, _args):
    return raw_nc.clns2am0_a0(z)


class AsyNNLLNonSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cns_NNLL_NNLO)


class PdfMatchingLLNonSinglet(EmptyPartonicChannel):
    pass


class PdfMatchingNLLNonSinglet(EmptyPartonicChannel):
    pass


class PdfMatchingNNLLNonSinglet(EmptyPartonicChannel):
    pass


class LightNonSingletShifted(pc.PartonicChannelAsy):
    def NNLO(self):
        return light.NonSinglet(self.ESF, self.nf).NLO()


class MatchingIntrinsicSplus(pc.FMatchingQuark):
    ffns = intrinsic.Splus


class MatchingIntrinsicSminus(pc.FMatchingQuark):
    ffns = intrinsic.Sminus


class MatchingGluonSplus(pc.FMatchingGluon):
    ffns = intrinsic.Splus


class MatchingGluonSminus(pc.FMatchingGluon):
    ffns = intrinsic.Sminus



# -*- coding: utf-8 -*-
from eko import basis_rotation as br

from .. import kernels


def import_pc_module(kind, process):
    return kernels.import_local(kind, process, __name__)


def generate(esf, nf):
    """
    Collect the light coefficient functions

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nf : int
            number of light flavors

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    pcs = import_pc_module(kind, esf.process)
    if esf.process == "CC":
        weights_even = kernels.cc_weights_even(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[:nf], nf
        )
        ns_even = kernels.Kernel(weights_even["ns"], pcs.NonSingletEven(esf, nf))
        g = kernels.Kernel(weights_even["g"], pcs.Gluon(esf, nf))
        s = kernels.Kernel(weights_even["s"], pcs.Singlet(esf, nf))
        weights_odd = kernels.cc_weights_odd(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[:nf], nf
        )
        ns_odd = kernels.Kernel(weights_odd["ns"], pcs.NonSingletOdd(esf, nf))
        return (ns_even, g, s, ns_odd)
    weights = nc_weights(esf.info.coupling_constants, esf.Q2, kind, nf)
    ns = kernels.Kernel(weights["ns"], pcs.NonSinglet(esf, nf))
    g = kernels.Kernel(weights["g"], pcs.Gluon(esf, nf))
    s = kernels.Kernel(weights["s"], pcs.Singlet(esf, nf))
    return (ns, g, s)


def nc_weights(coupling_constants, Q2, kind, nf, skip_heavylight=False):
    """
    Compute light NC weights.

    Parameters
    ----------
        coupling_constants : CouplingConstants
            manager for coupling constants
        Q2 : float
            W virtuality
        kind : str
            structure function kind
        nf : int
            number of light flavors
        skip_heavylight : bool
            prevent the last quark to couple to the boson

    Returns
    -------
        weights : dict
            mapping pid -> weight for ns, g and s channel
    """
    # quark couplings
    tot_ch_sq = 0
    ns_partons = {}
    pids = range(1, nf + 1)
    for q in pids:
        # we don't want this quark to couple to the photon (because it is treated separately),
        # but still let it take part in the average
        if skip_heavylight and q == nf:
            continue
        if kind != "F3":
            w = coupling_constants.get_weight(
                q, Q2, "VV"
            ) + coupling_constants.get_weight(q, Q2, "AA")
        else:
            w = coupling_constants.get_weight(
                q, Q2, "VA"
            ) + coupling_constants.get_weight(q, Q2, "AV")
        ns_partons[q] = w
        ns_partons[-q] = w if kind != "F3" else -w
        tot_ch_sq += w
    # gluon coupling = charge average (omitting the *2/2)
    ch_av = tot_ch_sq / len(pids) if kind != "F3" else 0.0
    # same for singlet
    s_partons = {q: ch_av for q in [*pids, *(-q for q in pids)]}
    return {"ns": ns_partons, "g": {21: ch_av}, "s": s_partons}



# -*- coding: utf-8 -*-
"""Coupling between QCD particles and EW particles."""
import logging

import numpy as np
from eko import basis_rotation as br

logger = logging.getLogger(__name__)


class CouplingConstants:
    """Defines the coupling constants between the QCD particles and the EW particles."""

    def __init__(self, theory_config, obs_config):
        """Create from dictionaries.

        Parameters
        ----------
        theory_config : dict
            theory settings
        obs_config : dic
            observable settings
        """
        self.theory_config = theory_config
        self.obs_config = obs_config
        # electric charges ----------------------------------------------------
        # QCD particles
        self.electric_charge = {21: 0}
        for q in range(1, 7):
            self.electric_charge[q] = 2 / 3 if q % 2 == 0 else -1 / 3
        # leptons: 11 = e-(!)
        for pid in [11, 13, 15]:
            self.electric_charge[pid] = -1
        # neutrinos
        for pid in [12, 14, 16]:
            self.electric_charge[pid] = 0

        # 3rd component of weak isospin ---------------------------------------
        # QCD particles
        self.weak_isospin_3 = {21: 0}
        for q in range(1, 7):
            self.weak_isospin_3[q] = 1 / 2 if q % 2 == 0 else -1 / 2  # u if stmt else d
        # leptons: 11 = e-(!)
        for pid in [11, 13, 15]:
            self.weak_isospin_3[pid] = -1 / 2
        # neutrinos
        for pid in [12, 14, 16]:
            self.weak_isospin_3[pid] = 1 / 2
        self.log()

    def log(self):
        """Write current configuration to log."""
        logger.info(self.theory_config)
        logger.info(self.obs_config)

    def vectorial_coupling(self, pid):
        """Combine the vectorial coupling from electric and weak charges."""
        return (
            self.weak_isospin_3[pid]
            - 2.0 * self.electric_charge[pid] * self.theory_config["sin2theta_weak"]
        )

    def leptonic_coupling(self, mode, quark_coupling_type):
        """Compute the coupling of the boson to the lepton.

        Parameters
        ----------
        mode : str
            scattered bosons
        quark_coupling_type : str
            (axial-)vectorial/(axial-)vectorial coupling

        Returns
        -------
        float
            leptonic coupling
        """
        # for CC the polarisation are NOT part of the structure functions, but are accounted for on
        # the cross section level. In order to have a true-trivial LO coeficient function, return
        # here 2.
        if mode == "WW":
            return 2

        # now NC only ...
        projectile_pid = self.obs_config["projectilePID"]
        # correct projectile polarization
        pol = self.obs_config["polarization"]
        if (projectile_pid % 2 == 1 and projectile_pid > 0) or (
            projectile_pid % 2 == 0 and projectile_pid < 0
        ):
            pol *= -1
        # load Z coupling
        projectile_v = 0.0
        projectile_a = 0.0
        if mode in ["phZ", "ZZ"]:
            projectile_v = self.vectorial_coupling(abs(projectile_pid))
            projectile_a = self.weak_isospin_3[abs(projectile_pid)]
        # switch mode
        if mode == "phph":
            if quark_coupling_type in ["VV", "AA"]:
                return self.electric_charge[abs(projectile_pid)] ** 2
            else:
                return 0
        elif mode == "phZ":
            if quark_coupling_type in ["VV", "AA"]:
                return self.electric_charge[abs(projectile_pid)] * (
                    projectile_v + pol * projectile_a
                )
            else:
                return self.electric_charge[abs(projectile_pid)] * (
                    projectile_a + pol * projectile_v
                )
        elif mode == "ZZ":
            if quark_coupling_type in ["VV", "AA"]:
                return (
                    projectile_v**2
                    + projectile_a**2
                    + 2.0 * pol * projectile_v * projectile_a
                )
            else:
                return 2.0 * projectile_v * projectile_a + pol * (
                    projectile_v**2 + projectile_a**2
                )
        raise ValueError(f"Unknown mode: {mode}")

    def partonic_coupling(self, mode, pid, quark_coupling_type, cc_mask=None):
        """Compute the coupling of the boson to the parton.

        Parameters
        ----------
        mode : str
            scattered bosons
        pid : int
            parton identifier
        quark_coupling_type : str
            flag to distinguish for heavy quarks between vectorial and axial-vectorial
            coupling
        cc_mask : str
            observable flavor to determine the heavy flavour couplings in F3

        Returns
        -------
        float
            partonic coupling
        """
        # for quarks only the flavor does matter
        pid = abs(pid)
        if mode == "WW":
            return np.sum(self.theory_config["CKM"].masked(cc_mask)(pid))
        # load couplings
        qph = {
            "V": self.electric_charge[pid],
            "A": 0,  # axial coupling of the photon to the quark is not there of course
        }
        qZ = {"V": self.vectorial_coupling(pid), "A": self.weak_isospin_3[pid]}
        first = quark_coupling_type[0]
        second = quark_coupling_type[1]
        # switch mode
        if mode == "phph":
            return qph[first] * qph[second]
        elif mode == "phZ":
            return qph[first] * qZ[second]
        elif mode == "ZZ":
            return qZ[first] * qZ[second]
        raise ValueError(f"Unknown mode: {mode}")

    def propagator_factor(self, mode, Q2):
        r"""Compute propagator correction to account for different bosons (:math:`\\eta` in PDG).

        Parameters
        ----------
        mode : str
            scattered bosons
        Q2 : float
            virtuality of the process

        Returns
        -------
        float
            propagator shift
        """
        if mode == "phph":
            return 1
        eta_phZ = (Q2 / (self.theory_config["MZ2"] + Q2)) / (
            4.0
            * self.theory_config["sin2theta_weak"]
            * (1.0 - self.theory_config["sin2theta_weak"])
        )
        eta_phZ /= 1 - self.obs_config["propagatorCorrection"]
        if mode == "phZ":
            return eta_phZ
        if mode == "ZZ":
            return eta_phZ**2
        if mode == "WW":
            eta_W = (
                (eta_phZ / 2)
                * (1 + Q2 / self.theory_config["MZ2"])
                / (1 + Q2 / self.theory_config["MW2"])
            ) ** 2
            return eta_W
        raise ValueError(f"Unknown mode: {mode}")

    def get_weight(self, pid, Q2, quark_coupling_type, cc_mask=None):
        """Compute the weight for the pid contributions to the structure function.

        Combine the charges, both on the leptonic side and the hadronic side, as well
        as propagator changes and/or corrections.

        Parameters
        ----------
        pid : int
            particle identifier
        Q2 : float
            DIS virtuality
        quark_coupling_type : str
            flag to distinguish for heavy quarks between vectorial and axial-vectorial
            coupling
        cc_mask : str
            observable flavor to determine the heavy flavour couplings in F3

        Returns
        -------
        float
            weight
        """
        # CC = WW
        if self.obs_config["process"] == "CC":
            return self.leptonic_coupling(
                "WW", quark_coupling_type
            ) * self.partonic_coupling("WW", pid, quark_coupling_type, cc_mask=cc_mask)
        # NC or EM
        # are we in positivity mode?
        # Note that for yadism the values None and "all" are equivalent
        # (they both mean no restriction on the couplings),
        # but this might be not the case for the caller (e.g. pinefarm)
        # since he may want to do further adjustments (such as deactivate TMC
        # in the theory).
        if (
            self.obs_config["nc_pos_charge"] is not None
            and self.obs_config["nc_pos_charge"] != "all"
        ):
            pos = self.obs_config["nc_pos_charge"][0]
            pos_pid = 1 + br.quark_names.index(pos)
            if abs(pid) != pos_pid:
                return 0.0

        w_phph = (
            self.leptonic_coupling("phph", quark_coupling_type)
            * self.propagator_factor("phph", Q2)
            * self.partonic_coupling("phph", pid, quark_coupling_type)
        )
        # pure photon exchane
        if self.obs_config["process"] == "EM":
            return w_phph
        # allow Z to be mixed in
        if self.obs_config["process"] == "NC":
            # photon-Z interference
            w_phZ = (
                2
                * self.leptonic_coupling("phZ", quark_coupling_type)
                * self.propagator_factor("phZ", Q2)
                * self.partonic_coupling("phZ", pid, quark_coupling_type)
            )
            # true Z contributions
            w_ZZ = (
                self.leptonic_coupling("ZZ", quark_coupling_type)
                * self.propagator_factor("ZZ", Q2)
                * self.partonic_coupling("ZZ", pid, quark_coupling_type)
            )
            return w_phph + w_phZ + w_ZZ
        raise ValueError(f"Unknown process: {self.obs_config['process']}")

    @classmethod
    def from_dict(cls, theory, observables):
        """Create the object from the theory dictionary.

        Parameters
        ----------
        theory : dict
            theory dictionary
        observables : dict
            observables dictionary

        Returns
        -------
        CouplingConstants
            created object
        """
        theory_config = {
            "MZ2": theory.get("MZ", 91.1876)
            ** 2,  # TODO remove defaults to the PDG2020 value
            "CKM": CKM2Matrix.from_str(
                theory["CKM"]
            ),  # TODO remove default in PDG2020 Eq. 12.33
            "sin2theta_weak": theory.get(
                "SIN2TW", 0.23121
            ),  # TODO remove defaults to the PDG2020 value
        }
        # set MW
        MW = theory.get("MW")
        if MW is None:
            theory_config["MW2"] = theory_config["MZ2"] / (
                1 - theory_config["sin2theta_weak"]
            )
            # TODO raise warning in log if inconsitent
        else:
            theory_config["MW2"] = MW**2

        # map projectile to PID
        proj = observables.get("ProjectileDIS", "electron")
        projectile_pids = {
            "electron": 11,
            "positron": -11,
            "neutrino": 12,
            "antineutrino": -12,
        }
        if proj not in projectile_pids:
            raise ValueError(f"Unknown projectile {proj}")
        obs_config = {
            "process": observables["prDIS"],
            "projectilePID": projectile_pids[proj],
            "polarization": observables["PolarizationDIS"],
            "propagatorCorrection": observables["PropagatorCorrection"],
            "nc_pos_charge": observables["NCPositivityCharge"],
        }
        o = cls(theory_config, obs_config)
        return o


class CKM2Matrix:
    """Wrapper for the CKM matrix."""

    flav_rows = ["u", "c", "t"]
    pid_rows = [2, 4, 6]
    flav_cols = ["d", "s", "b"]
    pid_cols = [1, 3, 5]

    def __init__(self, elems):
        """Create from matrix.

        Parameters
        ----------
            elems : list(float)
                squared elements in row order
        """
        self.m = np.array(elems).reshape(3, 3)
        # TODO maybe raise warning if non-unitarian

    def __repr__(self):
        """Return string representation."""
        return "CKM(" + str(self.m).replace("\n", "") + ")"

    def __getitem__(self, key):
        """Allow pid and strings as key.

        Parameters
        ----------
        key : int or str
            input key

        Returns
        -------
        float or list
            element(s)
        """
        nkey = []
        if not isinstance(key, tuple):
            key = (key,)
        if len(key) > 2:
            raise KeyError("CKM matrices are 3x3 matrices")
        for k, flavs, pids in zip(
            key, [self.flav_rows, self.flav_cols], [self.pid_rows, self.pid_cols]
        ):
            if isinstance(k, str):
                nkey.append(flavs.index(k))
            elif isinstance(k, int):
                nkey.append(pids.index(k))
            else:
                nkey.append(k)
        return self.m[tuple(nkey)]

    def __call__(self, pid):
        """Get column and row depending on pid.

        Parameters
        ----------
        pid : int
            particle identifier

        Returns
        -------
        list(float)
            row or column
        """
        if pid % 2 == 0:
            return self[pid]
        return self[:, pid]

    def masked(self, flavs):
        """Apply a mask according to the flavor.

        Parameters
        ----------
        flavs : str
            participating flavors as single characters

        Returns
        -------
        CKMMatrix
            masked matrix
        """
        op = np.zeros((3, 3))
        if "dus" in flavs:
            op += np.array([[1, 1, 0], [0, 0, 0], [0, 0, 0]])
        if "c" in flavs:
            op += np.array([[0, 0, 0], [1, 1, 0], [0, 0, 0]])
        if "b" in flavs:
            op += np.array([[0, 0, 1], [0, 0, 1], [0, 0, 0]])
        if "t" in flavs:
            op += np.array([[0, 0, 0], [0, 0, 0], [1, 1, 1]])
        return type(self)(self.m * op)

    @classmethod
    def from_str(cls, theory_string):
        """Create the object from a string representation.

        Parameters
        ----------
        theory_string : str
            all elements rowwise in a string


        Returns
        -------
        CKMMatrix
            created object
        """
        elems = theory_string.split(" ")
        return cls(np.power(np.array(elems, dtype=float), 2))



# -*- coding: utf-8 -*-

import numpy as np

from .. import splitting_functions as split
from ..partonic_channel import RSL
from . import partonic_channel as pc


class NonSinglet(pc.ChargedCurrentNonSinglet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sf_prefactor = 1.0 - self.labda

    def NLO(self):
        """
        |ref| implements :eqref:`A.1-4` with Table 1, :cite:`gluck-ccheavy`.
        """
        a = self.ka
        b1 = lambda z: (
            2.0 - 2.0 * z**2 - 2.0 / z - self.labda * (1.0 - 4.0 * z + z**2)
        )
        b2 = lambda z: 2.0 / z - 1.0 - z - self.labda * (z - z**2)

        return self.h_q(a, b1, b2)


class Gluon(pc.ChargedCurrentGluon):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sf_prefactor = 1.0 - self.labda

    def NLO(self):
        """
        |ref| implements :eqref:`A.5-7` with Table 2, :cite:`gluck-ccheavy`.
        """
        as_norm = 2.0

        def reg(z, _args):
            c1 = 8.0 * self.labda**2 - 6.0 * self.labda + 2.0
            c2 = 0.0
            c3 = 4.0 * self.labda
            c4 = -8.0 * self.labda
            return (
                (
                    self.sf_prefactor
                    * (split.lo.pqg_single(z, np.array([], dtype=float)) / 2.0)
                    * (self.l_labda(z) - np.log(self.labda))
                )
                + self.h_g(z, [c1, c2, c3, c4])
            ) * as_norm

        return RSL(reg)



# -*- coding: utf-8 -*-

import re
import textwrap


class DomainError(ValueError):
    """Error raised while checking inputs (single argument violation).

    .. todo::
        format domain properly according the type, this should be done by
        error Argument subclasses before raising.

    Parameters
    ----------
    name : str
        Name of the input variable.
    description : str
        Domain description provided.
    domain : list, optional
        The set of rules used to provide .
    """

    def __init__(
        self, *, name, description, type, known_as, value, domain_="", **kwargs
    ):
        """Generates the error message to be included in the Traceback.

        It formats the provided fields in a single string, and stores it as
        first element of ``args`` attribute.
        """
        self.name = name
        self.known_as = known_as
        msg = f"""Following argument outside the domain:
               name: {name}
               known_as: {known_as}
               description:
               DESCRIPTION
               type: {type}
               domain:
               DOMAIN
               value provided: {value}
               """
        domain = "\t" + re.sub("\n", "\n\t\t", domain_)
        description = "\n\t\t".join(
            textwrap.wrap(description.strip(), 66, break_long_words=False)
        )
        msg = re.sub("\n *", "\n\t", msg)
        msg = re.sub("DOMAIN", domain, msg)
        msg = re.sub("DESCRIPTION", f"\t{description}", msg)
        msg = re.sub("\t", "   ", msg)
        self.args = (msg,)
        # raise ValueError


class CrossConstraintError(ValueError):
    """Error raised while checking inputs (more arguments involved)."""

    def __init__(self):
        """Generate the error message to be included in the Traceback.

        It formats the provided fields in a single string, and stores it as
        first element of ``args`` attribute.

        ..todo::
            * define the arguments involved
            * define the actual format
        """
        pass


class DefaultError(ValueError):
    """Missing argument for field without default"""

    def __init__(self, default):
        """Generate the error message to be reported."""
        self.name = default["name"]
        msg = f"Missing default for the variable `{default['involve']}`"
        self.args = (msg,)


class DefaultWarning(RuntimeWarning):
    """Warn the user that a default is being applied"""

    def __init__(self, msg):
        """Generate the error message to be reported."""

        self.args = (msg,)



# -*- coding: utf-8 -*-
import numpy as np
from banana.data import cartesian_product, sql
from eko import interpolation

from . import db

default_card = dict(
    interpolation_xgrid=interpolation.make_grid(30, 20).tolist(),
    interpolation_polynomial_degree=4,
    interpolation_is_log=True,
    prDIS="EM",
    ProjectileDIS="electron",
    PolarizationDIS=0,
    PropagatorCorrection=0,
    TargetDIS="proton",
    NCPositivityCharge=None,
    observables={},
)
default_card = dict(sorted(default_card.items()))

default_kinematics = []
default_kinematics.extend(
    [dict(x=x, Q2=10.0, y=0) for x in default_card["interpolation_xgrid"][12::3]]
)
default_kinematics.extend(
    [dict(x=0.001, Q2=Q2, y=0) for Q2 in np.geomspace(4, 22, 10).tolist()]
)

default_config = {
    0: {"observable_names": ["F2_light"], "kinematics": default_kinematics},
    1: {
        "observable_names": [
            "F2_light",
            "F2_charm",
            "FL_light",
            "FL_charm",
            "F3_light",
            "F3_charm",
        ],
        "kinematics": default_kinematics,
    },
}

fns_config = {
    "ZM-VFNS": {
        "observable_names": ["F2_light", "FL_light", "F3_light"],
        "kinematics": default_kinematics,
    }
}


def build(observable_names, kinematics, update=None):
    """
    Generate all observable card updates

    Parameters
    ----------
        observable_names : list(str)
            observable names
        kinematics : list(dict)
            kinematics list
        update : dict
            base modifiers

    Returns
    -------
        cards : list(dict)
            list of update
    """
    cards = []
    if update is None:
        update = {}
    for c in cartesian_product(update):
        card = dict(observables={})
        card.update(c)
        for obs in observable_names:
            card["observables"][obs] = kinematics
        cards.append(card)
    return cards


# db interface
def load(session, updates):
    """
    Load observable records from the DB.

    Parameters
    ----------
        session : sqlalchemy.session.Session
            DB ORM session
        updates : dict
            modifiers

    Returns
    -------
        cards : list(dict)
            list of records
    """
    # add hash
    raw_records, df = sql.prepare_records(default_card, updates)
    # insert new ones
    sql.insertnew(session, db.Observable, df)
    return raw_records



# -*- coding: utf-8 -*-
import numpy as np
from eko import constants

from ..fonll import partonic_channel as pc
from ..partonic_channel import RSL, PartonicChannel
from . import raw_nc


class NeutralCurrentBase(pc.PartonicChannelAsyIntrinsic):
    def __init__(self, *args, m1sq, m2sq):
        super().__init__(*args, m1sq=m1sq, m2sq=m2sq)
        self.sigma_pp = self.Q2 + self.m2sq + self.m1sq
        self.sigma_mp = self.Q2 - self.m2sq + self.m1sq

    def init_nlo_vars(self):
        self.I1 = raw_nc.I1(self)
        self.Cplus = raw_nc.Cplus(self)
        self.C1m = raw_nc.C1m(self)
        self.C1p = raw_nc.C1p(self)
        self.CRm = raw_nc.CRm(self)
        self.S = raw_nc.S(self)
        self.L_xisoft = np.log(
            (self.sigma_pp - self.delta) / (self.sigma_pp + self.delta)
        )

    def init_vars(self, z):
        self.s1hat = (
            (1.0 - z)
            * ((self.delta - self.sigma_pm) * z + self.delta + self.sigma_pm)
            / 2.0
            / z
        )
        self.deltap = self.kinematic_delta(self.m1sq, self.s1hat + self.m2sq, -self.Q2)
        self.L_xi = np.log(
            (self.sigma_pp + self.s1hat - self.deltap)
            / (self.sigma_pp + self.s1hat + self.deltap)
        )

    def mkNLO(self, kind, RS):
        self.init_nlo_vars()
        norm = 2.0 * constants.CF * self.eta / self.x  # 2 = as_norm
        omx = norm * raw_nc.__getattribute__(  # pylint: disable=no-member
            f"f{kind}_{RS}_soft"
        )(self)

        delta = norm * (
            raw_nc.__getattribute__(f"f{kind}_{RS}_virt")(  # pylint: disable=no-member
                self
            )
            + self.S  # add normalization between curly and upright F
            * raw_nc.__getattribute__(f"m{kind}_{RS}")(  # pylint: disable=no-member
                self
            )
        )

        def reg(z, _args):
            self.init_vars(z)
            return norm * raw_nc.__getattribute__(  # pylint: disable=no-member
                f"f{kind}_{RS}_raw"
            )(self) - omx / (1.0 - z)

        return RSL.from_distr_coeffs(reg, (delta, omx))


class ChargedCurrentBase(PartonicChannel):
    """
    The convolution point simplifies to :math:`x` when m2=0,
    see :eqref:`6` of :cite:`kretzer-schienbein`.
    """

    def __init__(self, *args, m1sq):
        super().__init__(*args)
        self.m1sq = m1sq
        self.y = -self.ESF.Q2 / m1sq



# -*- coding: utf-8 -*-
from ...partonic_channel import RSL
from . import convolutions as conv
from . import non_singlet as ns
from . import singlet as s


def pgq0(_nf):
    """
    |LO| gluon-quark splitting function :math:`P_{gq}^{(0)}`.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction

    Returns
    -------
        float
            |LO| gluon-quark splitting function :math:`P_{gq}^{(0)}(z)`

    """
    return RSL(s.pgq0_reg)


def pgg0(nf):
    """
    |LO| gluon-gluon splitting function :math:`P_{gg}^{(0)}`.

    |ref| implements :eqref:`4.6`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        nf : int
            the number of light flavors

    Returns
    -------
        float
            |LO| gluon-gluon splitting function :math:`P_{gg}^{(0)}(z)`

    """
    return RSL(s.pgg0_reg, s.pgg0_sing, s.pgg0_local, [nf])


def pqq1(nf):
    """
    |NLO| quark-quark splitting function :math:`P_{qq}^{(1)}`.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        nf : int
            the number of light flavors

    Returns
    -------
        float
            |NLO| quark-quark splitting function :math:`P_{qq}^{(1)}(z)`

    """
    return RSL(s.pqq1_reg, ns.pns_sing, ns.pns_loc, [nf])


def pqg1(nf):
    """
    |NLO| quark-gluon splitting function :math:`P_{qg}^{(1)}`.

    |ref| implements :eqref:`4.8`, :cite:`split-singlet`.

    Parameters
    ----------
        z : float
            partonic momentum fraction
        nf : int
            the number of light flavors

    Returns
    -------
        float
            |NLO| quark-gluon splitting function :math:`P_{qg}^{(1)}(z)`

    """
    return RSL(s.pqg1_reg, args=[nf])


def pnsp1(nf):
    return RSL(ns.pnsp_reg, ns.pns_sing, ns.pns_loc, [nf])


def pnsm1(nf):
    return RSL(ns.pnsm_reg, ns.pns_sing, ns.pns_loc, [nf])


def pqq0_2(_nf):
    return RSL(conv.pqq0_2_reg, conv.pqq0_2_sing, conv.pqq0_2_loc)


def pqg0pgq0(nf):
    return RSL(conv.pqg0pgq0_reg, args=[nf])


def pqq0pqg0(nf):
    return RSL(conv.pqq0pqg0_reg, args=[nf])


def pqg0pgg0(nf):
    return RSL(conv.pqg0pgg0_reg, args=[nf])


raw_labels = {
    "P_gq_0": pgq0,
    "P_gg_0": pgg0,
    "P_qq_1": pqq1,
    "P_qg_1": pqg1,
    "P_nsp_1": pnsp1,
    "P_nsm_1": pnsm1,
    "P_qq_0^2": pqq0_2,
    "P_qg_0P_gq_0": pqg0pgq0,
    "P_qq_0P_qg_0": pqq0pqg0,
    "P_qg_0P_gg_0": pqg0pgg0,
}



# -*- coding: utf-8 -*-
import numba as nb
import numpy as np
from eko.constants import CF, TR


@nb.njit("f8(f8,f8[:])", cache=True)
def ns_reg(z, args):
    return CF * 4.0 * z


@nb.njit("f8(f8,f8[:])", cache=True)
def gluon_reg(z, args):
    nf = args[0]
    return nf * TR * 16 * z * (1.0 - z)



# -*- coding: utf-8 -*-
def ffns(nf):
    theory = {}

    for i, q in enumerate("cbt"):
        theory[f"ZM{q}"] = True
        theory[f"kDIS{q}Thr"] = 0.0 if nf >= i + 4 else float("inf")

    return theory



# -*- coding: utf-8 -*-
import logging
import time

import numpy as np
from eko import basis_rotation as br
from eko import beta
from scipy.special import binom

from ..coefficient_functions import splitting_functions as split
from .conv import convolute_operator

logger = logging.getLogger(__name__)


def build_orders(order):
    """
    Compute all necessary order configurations for a given power of alpha.

    Parameters
    ----------
        order : int
            order in pQCD

    Returns
    -------
        orders : list(tuple)
            order configurations
    """
    orders = []
    for alphas_power in range(order + 1):
        for lnf_power in range(alphas_power + 1):
            for lnrf_power in range(max(alphas_power, 1)):
                orders.append((alphas_power, 0, lnrf_power, lnf_power))
    return orders


class ScaleVariations:
    """
    Apply scale variations.

    Parameters
    ----------
        order : int
            perturbative order
        interpolator : eko.interpolation.InterpolationDispatcher
            interpolation basis functions
    """

    def __init__(self, order, interpolator, activate_ren, activate_fact):
        self.order = order
        self.interpolator = interpolator
        self.activate_ren = activate_ren
        self.activate_fact = activate_fact
        self.operators = {}
        self.raw_labels = split.raw_labels[: self.order]
        logger.info(
            "RenScaleVar: %s, FactScaleVar: %s", self.activate_ren, self.activate_fact
        )

    def compute_raw(self, nf):
        """
        Compute all basic building blocks.

        Parameters
        ----------
            nf : int
                number of active flavors
        """
        # compute all raw ingredients
        for order_labels in self.raw_labels:
            for l, fnc in order_labels.items():
                if (l, nf) in self.operators:
                    logger.debug("using cached %s", l)
                    continue
                start_time = time.perf_counter()
                # TODO add error propagation
                res, _err = convolute_operator(fnc(nf), self.interpolator)
                self.operators[(l, nf)] = res
                logger.info(
                    "computing %s - took: %f s", l, time.perf_counter() - start_time
                )

    def fact_matrices(self, nf):
        r"""
        Compute all matrices related to factorization scale variation,
        i.e. :math:`\ln(Q^2/\mu_F^2)`.

        Parameters
        ----------
            nf : int
                number of active flavors

        Returns
        -------
            dict :
                map with `(target, lnf, src) -> np.ndarray`
        """
        self.compute_raw(nf)
        # load mappings
        smap = split.sector_mapping(self.order, self.operators, nf)
        fact_matrices = {}
        # collect all elements
        for k, sectors in smap.items():
            # propagete sectors
            fact_op = []
            for ad in br.anomalous_dimensions_basis:
                fact_op.append(sectors[ad])
            fact_matrices[k] = np.array(fact_op)

        return fact_matrices

    def ren_coeffs(self, nf):
        r"""
        Provide the renormalization scale variation coefficients,
        i.e. :math:`\ln(\mu_F^2/\mu_R^2)`.

        Parameters
        ----------
            nf : int
                number of active flavors

        Returns
        -------
            dict :
                map with `(target, lnf2r, src) -> np.ndarray`

        """
        ren_coeffs = {(2, 1, 1): -beta.beta_0(nf)}
        return dict(filter(lambda item: item[0][0] <= self.order, ren_coeffs.items()))

    def apply_common_scale_variations(self, ker_orders, nf):
        """
        Add new kernels for common scale varied coefficient functions.

        Parameters
        ----------
            ker_orders : list
                raw (unscale-varied) coefficient functions
            nf : int
                number of active flavors

        Returns
        -------
            list :
                kernels map
        """
        if not self.activate_fact:
            return []
        # get the two ingredients: matrices and projectors
        fmatrices = self.fact_matrices(nf)
        projectors = br.ad_projectors(nf)
        # join everything together
        added_ker_sv = []
        for (o, oqed, _, _), ker in ker_orders:
            partons_proj = ker[0][:, 0] @ projectors
            for (target, lnf, src), fmat in fmatrices.items():
                if src == o:
                    val_sv = fmat @ ker[1][0]
                    err_sv = fmat @ ker[2][0]
                    added_ker_sv.append(
                        ((target, oqed, 0, lnf), (partons_proj.T, val_sv, err_sv))
                    )
        return added_ker_sv

    def apply_diff_scale_variations(self, ker_orders, nf):
        """
        Add new kernels for different scale varied coefficient functions.

        Parameters
        ----------
            ker_orders : list
                common-scale varied coefficient functions
            nf : int
                number of active flavors

        Returns
        -------
            Iterable :
                kernels map
        """
        # if none is active - perfect, nothing to do
        if not self.activate_fact and not self.activate_ren:
            return []
        diff_kers = self.apply_raw_diff_scale_variations(ker_orders, nf)
        ren_kers = []
        for (o, k) in diff_kers:
            # ln((xi_f/xi_r)^2)^n = (ln(xi_f^2) - ln(xi_r^2))^n
            n = o[2]
            for j in range(
                n + 1
            ):  # j is the power of ln(xi_r^2), n-j is the power of ln(xi_f^2)
                binomial = binom(n, j) * (-1) ** j  # take care of the additional minus
                ren_kers.append(
                    ((o[0], o[1], j, n - j + o[3]), (binomial * k[0], k[1], k[2]))
                )
        # if at least one was trivial, skip that one (since the user asked for it)
        if not self.activate_ren:
            return filter(lambda e: e[0][2] == 0, ren_kers)
        if not self.activate_fact:
            return filter(lambda e: e[0][3] == 0, ren_kers)
        # ok, we need to return the full thing
        return ren_kers

    def apply_raw_diff_scale_variations(self, ker_orders, nf):
        """
        Add new kernels for different scale varied coefficient functions.

        Parameters
        ----------
            ker_orders : list
                common-scale varied coefficient functions
            nf : int
                number of active flavors

        Returns
        -------
            list :
                kernels map
        """
        ren_coeffs = self.ren_coeffs(nf)
        # join everything together
        added_ker_sv = []
        for (o, oqed, _, lnf), ker in ker_orders:
            # TODO: APFEL is wrong - fix it temporarily here
            # if (o, lnf) == (1, 1):
            #     continue
            for (target, lnf2r, src), rcoeff in ren_coeffs.items():
                if src == o:
                    added_ker_sv.append(
                        ((target, oqed, lnf2r, lnf), (rcoeff * ker[0], ker[1], ker[2]))
                    )
        return added_ker_sv



# -*- coding: utf-8 -*-
from ..partonic_channel import RSL
from . import f2_nc, nnlo


class NonSingletEven(f2_nc.NonSinglet):
    pass


class Gluon(f2_nc.Gluon):
    pass


class Singlet(f2_nc.Singlet):
    pass


class NonSingletOdd(f2_nc.NonSinglet):
    def NNLO(self):
        """
        |ref| implements :eqref:`2.9`, :cite:`vogt-f2lcc`.
        """

        return RSL(
            nnlo.xc2ns2p.c2nc2a, nnlo.xc2ns2p.c2ns2b, nnlo.xc2ns2p.c2nc2c, [self.nf]
        )



# -*- coding: utf-8 -*-
import banana.cfg
import numpy as np
import pandas as pd
from banana.benchmark.runner import BenchmarkRunner
from banana.data import dfdict
from eko.strong_coupling import StrongCoupling

import yadism
from yadmark.data import db, observables


class Runner(BenchmarkRunner):
    alphas_from_lhapdf = False
    """Use the alpha_s routine provided by the Pdf?"""

    db_base_cls = db.Base

    @staticmethod
    def load_ocards(conn, ocard_updates):
        return observables.load(conn, ocard_updates)

    def run_me(self, theory, ocard, pdf):
        """
        Run yadism

        Parameters
        ----------
            theory : dict
                theory card
            ocard : dict
                observable card
            pdf : lhapdf_like
                PDF set

        Returns
        -------
            out : yadism.output.PDFOutput
                yadism output
        """
        runner = yadism.Runner(theory, ocard)
        # choose alpha_s source
        if self.alphas_from_lhapdf:
            import lhapdf  # pylint:disable=import-outside-toplevel

            alpha_s = lambda muR: lhapdf.mkAlphaS(pdf.set().name).alphasQ(muR)
        else:
            new_theory, _ = yadism.input.compatibility.update(theory, ocard)
            sc = StrongCoupling.from_dict(new_theory)
            alpha_s = lambda muR: sc.a_s(muR**2) * 4.0 * np.pi

        alpha_qed = lambda _muR: theory["alphaqed"]
        return runner.get_result().apply_pdf_alphas_alphaqed_xir_xif(
            pdf, alpha_s, alpha_qed, theory["XIR"], theory["XIF"]
        )

    def run_external(self, theory, ocard, pdf):
        """
        Run yadism

        Parameters
        ----------
            theory : dict
                theory card
            ocard : dict
                observable card
            pdf : lhapdf_like
                PDF set

        Returns
        -------
            dict
                external output
        """
        observable = ocard

        if self.external.upper() == "APFEL":
            from .external import (  # pylint:disable=import-error,import-outside-toplevel
                apfel_utils,
            )

            return apfel_utils.compute_apfel_data(theory, observable, pdf)

        elif self.external.upper() == "QCDNUM":
            from .external import (  # pylint:disable=import-error,import-outside-toplevel
                qcdnum_utils,
            )

            return qcdnum_utils.compute_qcdnum_data(theory, observable, pdf)

        elif self.external.lower() == "xspace_bench":
            from .external import (  # pylint:disable=import-error,import-outside-toplevel
                xspace_bench_utils,
            )

            return xspace_bench_utils.compute_xspace_bench_data(theory, observable, pdf)

        elif self.external.lower() == "void":
            # set all ESF simply to 0
            res = {}
            for sf, esfs in ocard["observables"].items():
                if not yadism.observable_name.ObservableName.is_valid(sf):
                    continue
                void_esfs = []
                for esf in esfs:
                    n = esf.copy()
                    n.update({"result": 0})
                    void_esfs.append(n)
                res[sf] = void_esfs
            return res
        raise ValueError("Unknown external")

    def log(self, t, o, _pdf, me, ext):
        log_tab = dfdict.DFdict()
        kins = ["x", "Q2"]

        for sf in me:
            if not yadism.observable_name.ObservableName.is_valid(sf):
                continue
            esfs = []

            obs_kins = kins.copy()

            # skip if it is empty
            if len(me[sf]) == 0:
                continue

            if "y" in me[sf][0]:
                obs_kins += ["y"]

            # Sort the point using yadism order since yadism list can be different from ext
            for yad in me[sf]:
                cnt = 0
                for oth in ext[sf]:
                    if all([yad[k] == oth[k] for k in obs_kins]):
                        # add common values
                        esf = {}
                        esf["x"] = yad["x"]
                        esf["Q2"] = yad["Q2"]
                        if "y" in obs_kins:
                            esf["y"] = yad["y"]
                        esf["yadism"] = f = yad["result"]
                        esf["yadism_error"] = yad["error"]
                        esf[self.external] = r = oth["result"]
                        esf["percent_error"] = (f - r) / r * 100
                        esfs.append(esf)
                        cnt = 1
                        break
                if cnt == 0:
                    raise ValueError("Sort problem: x and/or Q2 do not match.")
            log_tab[sf] = pd.DataFrame(esfs)

        return log_tab



# -*- coding: utf-8 -*-
from ..intrinsic import f3_nc as intrinsic
from ..light import f3_nc as light
from ..partonic_channel import EmptyPartonicChannel
from . import partonic_channel as pc


class PdfMatchingLLNonSinglet(pc.PdfMatchingLLNonSinglet):
    pass


class PdfMatchingNLLNonSinglet(pc.PdfMatchingNLLNonSinglet):
    pass


class PdfMatchingNNLLNonSinglet(pc.PdfMatchingNNLLNonSinglet):
    pass


class LightNonSingletShifted(pc.PartonicChannelAsy):
    def NNLO(self):
        return light.NonSinglet(self.ESF, self.nf).NLO()


class MatchingIntrinsicRplus(pc.FMatchingQuark):
    ffns = intrinsic.Rplus


class MatchingIntrinsicRminus(pc.FMatchingQuark):
    ffns = intrinsic.Rminus


# TODO extract coefficient function from Buza/LeProHQ: see Eq. B.4 of Nucl. Phys. B485-420
# and remember that in the non-singlet case d_{2xg1} = d_{xF3}
class AsyLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNLLNonSinglet(EmptyPartonicChannel):
    pass


class AsyNNLLNonSinglet(EmptyPartonicChannel):
    pass



# -*- coding: utf-8 -*-
from banana.data.db import Base
from sqlalchemy import Column, Integer, Text


class Observable(Base):
    __tablename__ = "observables"

    PolarizationDIS = Column(Integer)
    ProjectileDIS = Column(Text)
    PropagatorCorrection = Column(Integer)
    interpolation_is_log = Column(Text)
    interpolation_polynomial_degree = Column(Integer)
    interpolation_xgrid = Column(Text)
    observables = Column(Text)
    prDIS = Column(Text)
    TargetDIS = Column(Text)
    NCPositivityCharge = Column(Text)



# -*- coding: utf-8 -*-

import numba as nb

from ..intrinsic import f2_nc as intrinsic
from ..light import f2_nc as light
from ..partonic_channel import RSL
from . import partonic_channel as pc
from . import raw_nc


@nb.njit("f8(f8,f8[:])", cache=True)
def cg_LL_NLO(z, args):
    L = args[0]
    return raw_nc.c2g1am0_aq(z) * L


@nb.njit("f8(f8,f8[:])", cache=True)
def cg_NLL_NLO(z, _args):
    return raw_nc.c2g1am0_a0(z)


@nb.njit("f8(f8,f8[:])", cache=True)
def cg_LL_NNLO(z, args):
    L = args[0]
    return (raw_nc.c2g2am0_aq2(z) - raw_nc.c2g2am0_aqf(z)) * L**2


@nb.njit("f8(f8,f8[:])", cache=True)
def cg_NLL_NNLO(z, args):
    L = args[0]
    return (raw_nc.c2g2am0_aq(z) - raw_nc.c2g2am0_af(z)) * L


@nb.njit("f8(f8,f8[:])", cache=True)
def cg_NNLL_NNLO(z, _args):
    return raw_nc.c2g2am0_a0(z)


class AsyLLGluon(pc.PartonicChannelAsy):
    def NLO(self):
        return RSL(cg_LL_NLO, args=[self.L])

    def NNLO(self):
        return RSL(cg_LL_NNLO, args=[self.L])


class AsyNLLGluon(pc.PartonicChannelAsy):
    def NLO(self):
        return RSL(cg_NLL_NLO, args=[self.L])

    def NNLO(self):
        return RSL(cg_NLL_NNLO, args=[self.L])


class AsyNNLLGluon(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cg_NNLL_NNLO)


@nb.njit("f8(f8,f8[:])", cache=True)
def cps_LL_NNLO(z, args):
    L = args[0]
    return (raw_nc.c2ps2am0_aq2(z) - raw_nc.c2ps2am0_aqf(z)) * L**2


@nb.njit("f8(f8,f8[:])", cache=True)
def cps_NLL_NNLO(z, args):
    L = args[0]
    return (raw_nc.c2ps2am0_aq(z) - raw_nc.c2ps2am0_af(z)) * L


@nb.njit("f8(f8,f8[:])", cache=True)
def cps_NNLL_NNLO(z, _args):
    return raw_nc.c2ps2am0_a0(z)


class AsyLLSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cps_LL_NNLO, args=[self.L])


class AsyNLLSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cps_NLL_NNLO, args=[self.L])


class AsyNNLLSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cps_NNLL_NNLO)


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_LL_NNLO_reg(z, args):
    L = args[0]
    return raw_nc.c2ns2am0_aq2(z) * L**2


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_LL_NNLO_sing(z, args):
    L = args[0]
    return raw_nc.c2ns2bm0_aq2(z) * L**2


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_LL_NNLO_loc(z, args):
    L = args[0]
    return raw_nc.c2ns2cm0_aq2(z) * L**2


class AsyLLNonSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cns_LL_NNLO_reg, cns_LL_NNLO_sing, cns_LL_NNLO_loc, args=[self.L])


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_NLL_NNLO_reg(z, args):
    L = args[0]
    return raw_nc.c2ns2am0_aq(z) * L


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_NLL_NNLO_sing(z, args):
    L = args[0]
    return raw_nc.c2ns2bm0_aq(z) * L


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_NLL_NNLO_loc(z, args):
    L = args[0]
    return raw_nc.c2ns2cm0_aq(z) * L


class AsyNLLNonSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cns_NLL_NNLO_reg, cns_NLL_NNLO_sing, cns_NLL_NNLO_loc, args=[self.L])


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_NNLL_NNLO_reg(z, _args):
    return raw_nc.c2ns2am0_a0(z)


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_NNLL_NNLO_sing(z, _args):
    return raw_nc.c2ns2bm0_a0(z)


@nb.njit("f8(f8,f8[:])", cache=True)
def cns_NNLL_NNLO_loc(z, _args):
    return raw_nc.c2ns2cm0_a0(z)


class AsyNNLLNonSinglet(pc.PartonicChannelAsy):
    def NNLO(self):
        return RSL(cns_NNLL_NNLO_reg, cns_NNLL_NNLO_sing, cns_NNLL_NNLO_loc)


class LightNonSingletShifted(pc.PartonicChannelAsy):
    def NNLO(self):
        return light.NonSinglet(self.ESF, self.nf).NLO()


class PdfMatchingLLNonSinglet(pc.PdfMatchingLLNonSinglet):
    pass


class PdfMatchingNLLNonSinglet(pc.PdfMatchingNLLNonSinglet):
    pass


class PdfMatchingNNLLNonSinglet(pc.PdfMatchingNNLLNonSinglet):
    pass


class MatchingIntrinsicSplus(pc.FMatchingQuark):
    ffns = intrinsic.Splus


class MatchingIntrinsicSminus(pc.FMatchingQuark):
    ffns = intrinsic.Sminus


class MatchingGluonSplus(pc.FMatchingGluon):
    ffns = intrinsic.Splus


class MatchingGluonSminus(pc.FMatchingGluon):
    ffns = intrinsic.Sminus



# -*- coding: utf-8 -*-
from .. import partonic_channel as pc
from ..partonic_channel import RSL
from . import f2_nc, nlo, nnlo


class NonSinglet(f2_nc.NonSinglet):
    @staticmethod
    def NLO():
        """
        |ref| implements :eqref:`155`, :cite:`moch-f3nc`.
        """

        return RSL.from_distr_coeffs(
            nlo.f3.ns_reg, (nlo.f2.ns_delta, nlo.f2.ns_omx, nlo.f2.ns_logomx)
        )

    def NNLO(self):
        """
        |ref| implements :eqref:`4.8`, :cite:`vogt-f2nc`.
        """

        return RSL(
            nnlo.xc3ns2p.c3np2a, nnlo.xc3ns2p.c3ns2b, nnlo.xc3ns2p.c3np2c, [self.nf]
        )


class Gluon(pc.EmptyPartonicChannel):
    pass


class Singlet(pc.EmptyPartonicChannel):
    pass



# -*- coding: utf-8 -*-
import argparse
import pathlib

import banana
from banana import cfg as banana_cfg
from banana import navigator as bnav

from . import glob, navigator

here = pathlib.Path(__file__).parent
banana.register(here.parents[2] / "benchmarks")


def yelp(*args):
    """
    Help function (renamed to avoid clash of names) - short cut: h.
    """
    if len(args) == 0:
        print(
            f"""Welcome to yadmark navigator - the yadism benchmark skript!
Available variables:
    {bnav.help_vars}
    o = "{bnav.o}" -> query observables
Available functions:
    {bnav.help_fncs}
    dfl(id) - log as DataFrame
    simlogs(id) - find similar logs
    diff(id,id) - subtractig logs
    compare(id,id) - compare externals
    check_log(id) - check logs passed
    crashed_log(id) - print crashed logs
"""
        )
    elif len(args) == 1:
        return help(*args)
    return None


def register_globals(configpath):
    app = navigator.NavigatorApp(configpath, "sandbox")
    glob.app = app

    glob.glob["yelp"] = yelp
    glob.glob["h"] = yelp

    # register banana functions
    bnav.register_globals(glob.glob, glob.app)

    # add my functions
    glob.glob["check_log"] = app.check_log


def launch_navigator():
    """CLI Entry point"""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-c", "--config", type=pathlib.Path, default=None, help="Path to config file"
    )

    args = parser.parse_args()

    register_globals(banana_cfg.detect(args.config))

    return bnav.launch_navigator(
        ["yadism", "yadmark", "yadmark.navigator.glob"], skip_cfg=True
    )



# -*- coding: utf-8 -*-

import numpy as np

hqfl = "cbt"


def update(theory, observables):
    """
    Upgrade the legacy theory and observable runcards with the new settings.

    Parameters
    ----------
        theory : dict
            theory runcard
        observables : dict
            observable runcard

    Returns
    -------
        new_theory : dict
            upgraded theory runcard
        new_obs : dict
            upgraded observable runcard
    """
    new_theory = theory.copy()
    new_obs = observables.copy()
    update_fns(new_theory)
    update_scale_variations(new_theory)
    update_target(new_obs)
    return new_theory, new_obs


def update_scale_variations(theory):
    """
    Extract necessity to compute scale variations.

    Parameters
    ----------
        theory : dict
            theory runcard
    """
    if "RenScaleVar" not in theory:
        theory["RenScaleVar"] = not np.isclose(theory["XIR"], 1.0)
    if "FactScaleVar" not in theory:
        theory["FactScaleVar"] = not np.isclose(theory["XIF"], 1.0)


def update_target(obs):
    """
    Map TargetDIS string to a (Z,A) dict.

    Parameters
    ----------
        obs : dict
            observable runcard
    """
    target = obs["TargetDIS"]
    if not isinstance(target, str):
        return
    if target == "proton":
        obs["TargetDIS"] = {"Z": 1.0, "A": 1.0}
        # proton = 2212
        obs["TargetDISid"] = "2212"
    elif target == "neutron":
        obs["TargetDIS"] = {"Z": 0.0, "A": 1.0}
        # neutron = 2112
        obs["TargetDISid"] = "2112"
    elif target == "isoscalar":
        obs["TargetDIS"] = {"Z": 1.0, "A": 2.0}
        # deuteron = 10 0 001 002 0
        obs["TargetDISid"] = "1000010020"
    elif target == "iron":
        obs["TargetDIS"] = {
            "Z": 23.403,
            "A": 49.618,
        }  # Fe=26 and we don't know how these factors got inside APFEL
        # iron = 10 0 026 056 0
        obs["TargetDISid"] = "1000260560"
    elif target == "lead":
        obs["TargetDIS"] = {"Z": 82.0, "A": 208.0}  # from hep-ex/0102049
        # lead = 10 0 082 208 0
        obs["TargetDISid"] = "1000822080"
    elif target == "neon":
        obs["TargetDIS"] = {"Z": 10.0, "A": 20.0}
        obs["TargetDISid"] = "1000100200"
    elif target == "marble":
        A = 40 + 3 * 16 + 12  # A(CaCO3): Atomic MAss
        Z = 20 + 3 * 8 + 6  # Z(CaCO3): Atomic Number
        obs["TargetDIS"] = {"Z": Z, "A": A}
        obs["TargetDISid"] = f"1000{str(Z)}{str(A)}0"
    else:
        raise ValueError(f"Unknown TargetDIS '{target}'")


def update_fns(theory):
    """
    Shifts kcThr around and add the kDIScThr parameter (likewise for all heavy quarks).

    Parameters
    ----------
        theory : dict
            theory runcard
    """
    fns = theory["FNS"]
    nf = theory["NfFF"]

    if "FONLL" in fns:
        theory["ZMc"] = False
        theory["ZMb"] = False
        theory["ZMt"] = True
    elif fns == "ZM-VFNS":
        for fl in hqfl:
            theory[f"ZM{fl}"] = True
    elif "FFNS" in fns:
        # enforce correct settings moving all thresholds to 0 or oo
        for k, fl in enumerate(hqfl):
            if k + 4 <= nf:
                theory[f"k{fl}Thr"] = 0.0
                theory[f"ZM{fl}"] = True
            else:
                theory[f"k{fl}Thr"] = np.inf
                theory[f"ZM{fl}"] = False
    else:
        raise ValueError(f"Scheme '{fns}' not recognized.")

    # here there is no difference between DGLAP and DIS
    for fl in hqfl:
        theory[f"kDIS{fl}Thr"] = theory[f"k{fl}Thr"]

    # fix FONLL-B and introduce PTODIS
    if fns == "FONLL-B":
        theory["PTODIS"] = 2
        theory["PTO"] = 1
    else:
        theory["PTODIS"] = theory["PTO"]



# -*- coding: utf-8 -*-
from eko import basis_rotation as br

from .. import kernels
from ..light.kernels import nc_weights as light_nc_weights


def import_pc_module(kind, process):
    return kernels.import_local(kind, process, __name__)


def generate(esf, nf, ihq):
    """
    Collect the heavy coefficient functions

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nf : int
            number of light flavors

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    pcs = import_pc_module(kind, esf.process)
    m2hq = esf.info.m2hq[ihq - 4]
    if esf.process == "CC":
        # TODO: use ihq for couplings
        w = kernels.cc_weights(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[nf], nf
        )
        return (
            kernels.Kernel(w["ns"], pcs.NonSinglet(esf, nf, m2hq=m2hq)),
            kernels.Kernel(w["g"], pcs.Gluon(esf, nf, m2hq=m2hq)),
        )
    else:
        # F3 is a non-singlet quantity and hence has neither gluon nor singlet-like contributions
        if kind == "F3":
            return ()
        weights = nc_weights(esf.info.coupling_constants, esf.Q2, kind, nf)
        gVV = kernels.Kernel(weights["gVV"], pcs.GluonVV(esf, nf, m2hq=m2hq))
        gAA = kernels.Kernel(weights["gAA"], pcs.GluonAA(esf, nf, m2hq=m2hq))
        sVV = kernels.Kernel(weights["sVV"], pcs.SingletVV(esf, nf, m2hq=m2hq))
        sAA = kernels.Kernel(weights["sAA"], pcs.SingletAA(esf, nf, m2hq=m2hq))
        return (gVV, gAA, sVV, sAA)


def generate_missing(esf, nf, ihq, icoupl=None):
    """
    Collect the missing coefficient functions

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nf : int
            number of light flavors
        ihq : int
            PID of heavy flavor
        icoupl : None or int
            PID of the flavor coupling (default: None)

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    pcs = import_pc_module(kind, esf.process)
    m2hq = esf.info.m2hq[ihq - 4]
    # in CC there are no missing diagrams known yet
    if esf.process == "CC":
        return ()
    weights = light_nc_weights(esf.info.coupling_constants, esf.Q2, kind, nf)
    if icoupl is not None:
        weights["ns"] = {k: v for k, v in weights["ns"].items() if abs(k) == icoupl}
    return (kernels.Kernel(weights["ns"], pcs.NonSinglet(esf, nf, m2hq=m2hq)),)


def nc_weights(coupling_constants, Q2, kind, nf):
    """
    Compute heavy NC weights.

    Parameters
    ----------
        coupling_constants : CouplingConstants
            manager for coupling constants
        Q2 : float
            boson virtuality
        kind : str
            structure function kind
        nf : int
            number of light flavors

    Returns
    -------
        weights : dict
            mapping pid -> weight
    """
    ihq = nf + 1
    if kind == "F3":
        return {}
    weight_vv = coupling_constants.get_weight(ihq, Q2, "VV")
    weight_aa = coupling_constants.get_weight(ihq, Q2, "AA")
    sVV = {}
    sAA = {}
    for q in range(1, nf + 1):
        sVV[q] = sVV[-q] = weight_vv
        sAA[q] = sAA[-q] = weight_aa
    return {"gVV": {21: weight_vv}, "gAA": {21: weight_aa}, "sVV": sVV, "sAA": sAA}



# -*- coding: utf-8 -*-
import numpy as np
from eko import basis_rotation as br
from eko.constants import TR

from .. import heavy, kernels, light


def import_pc_module(kind, process, subpkg=None):
    if subpkg is None:
        subpkg = __name__
    else:
        subpkg = ".".join(__name__.split(".")[:-2] + [subpkg, ""])
    return kernels.import_local(kind, process, subpkg)


def generate_light(esf, nl, pto_evol):
    r"""
    Collect the light coefficient functions for |FONLL|.

    |ref| implements :eqref:`96`, :cite:`forte-fonll`.

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nl : int
            number of light flavors
        pto_evol : int
            PTO of evolution

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    ihq = nl + 1
    # rewrite the derivative term back as a sum
    # and so we're back to cbar^{(nl)}
    light_elems = light.kernels.generate(esf, nl)
    kind = esf.info.obs_name.kind
    mu2hq = esf.info.threshold.area_walls[ihq - 3]
    L = np.log(esf.Q2 / mu2hq)
    fonll_cfs = import_pc_module(kind, esf.process)

    if esf.process == "CC":
        light_weights = kernels.cc_weights(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[:nl], nl
        )
    else:
        light_weights = light.kernels.nc_weights(
            esf.info.coupling_constants, esf.Q2, kind, nl
        )

    # Pdf matching conditions
    pdf_matching = []
    for res in range(pto_evol + 1):
        name = "PdfMatching" + ("N" * res) + "LL" + "NonSinglet"
        pdf_matching.append(
            -kernels.Kernel(
                light_weights["ns"],
                fonll_cfs.__getattribute__(name)(esf, nl, mu2hq=mu2hq),
            )
        )

    # alpha s matching condition
    as_norm = 2.0
    alphas_matching = []
    if pto_evol >= 1:  # contributes at NLL
        alphas_matching.append(
            -(2.0 / 3.0 * TR * L)
            * as_norm
            * kernels.Kernel(
                light_weights["ns"],
                fonll_cfs.LightNonSingletShifted(esf, nl, mu2hq=mu2hq),
            )
        )

    return (
        # true light contributions
        *light_elems,
        # matching
        *pdf_matching,
        *alphas_matching,
        # missing is dealt above in order to reuse this method in diff
    )


def generate_light_diff(esf, nl, pto_evol):
    r"""
    Collect the light diff coefficient functions for |FONLL|.

    |ref| implements :eqref:`95`, :cite:`forte-fonll`.

    Following the reference we have to collect the
    contributions of the *incoming* would-be-heavy quarks to the light
    structure function, where light means only non-heavy charges are active.
    So the incoming line can *not* be the one coupling to the photon, i.e. we
    are left with the singlet-like contributions.

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nl : int
            number of light flavors
        pto_evol : int
            PTO of evolution

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    light_cfs = import_pc_module(kind, esf.process, "light")
    if esf.process == "CC":
        light_weights = kernels.cc_weights(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[:nl], nl + 1
        )
    else:
        light_weights = light.kernels.nc_weights(
            esf.info.coupling_constants, esf.Q2, kind, nl + 1, skip_heavylight=True
        )
    s_w = {nl + 1: light_weights["s"][nl + 1], -(nl + 1): light_weights["s"][-(nl + 1)]}
    k = kernels.Kernel(s_w, light_cfs.Singlet(esf, nl + 1))
    k.max_order = pto_evol

    # the asy has all the light stuff again, so subtract it back
    asy = []
    light_elems = generate_light(esf, nl, pto_evol)
    for e in light_elems:
        e.min_order = pto_evol + 1
        asy.append(-e)
    # add in addition it also has the asymptotic limit of missing
    ihq = nl + 1
    mu2hq = esf.info.threshold.area_walls[ihq - 3]
    fonll_cfs = import_pc_module(kind, esf.process)
    for res in range(pto_evol + 1):
        name = "Asy" + ("N" * res) + "LL" + "NonSinglet"
        km = kernels.Kernel(
            light_weights["ns"], fonll_cfs.__getattribute__(name)(esf, nl, mu2hq=mu2hq)
        )
        asy.append(-km)
    return (k, *asy)


def generate_heavy_diff(esf, nl, pto_evol):
    """
    |ref| implements :eqref:`89`, :cite:`forte-fonll`.

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nl : int
            number of light flavors
        pto_evol : int
            PTO of evolution

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    ihq = nl + 1
    # add light contributions
    lights = kernels.generate_single_flavor_light(esf, nl + 1, ihq)
    for e in lights:
        e.max_order = pto_evol
    # add asymptotic contributions
    # The matching does not necessarily happen at the quark mass
    # m2hq = esf.info.m2hq[ihq - 4]
    # but will be done at the proper threshold
    fonll_cfs = import_pc_module(kind, esf.process)
    mu2hq = esf.info.threshold.area_walls[ihq - 3]
    asys = []
    if esf.process == "CC":
        wa = kernels.cc_weights(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[ihq - 1], nl
        )
        asys = [
            -kernels.Kernel(wa["ns"], fonll_cfs.AsyQuark(esf, nl, mu2hq=mu2hq)),
            -kernels.Kernel(wa["g"], fonll_cfs.AsyGluon(esf, nl, mu2hq=mu2hq)),
        ]
    else:
        asy_weights = heavy.kernels.nc_weights(
            esf.info.coupling_constants, esf.Q2, kind, nl
        )
        if kind != "F3":
            for c, channel in (("g", "Gluon"), ("s", "Singlet")):
                for res in range(pto_evol + 1):
                    name = "Asy" + ("N" * res) + "LL" + channel
                    for av in ("AA", "VV"):
                        asys.append(
                            -kernels.Kernel(
                                asy_weights[f"{c}{av}"],
                                fonll_cfs.__getattribute__(name)(esf, nl, mu2hq=mu2hq),
                            )
                        )

    return (*lights, *asys)


def generate_heavy_intrinsic_diff(esf, nl, pto_evol):
    """
    |ref| implements :eqref:`B.24-26`, :cite:`luca-intrinsic`.

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nl : int
            number of light flavors
        pto_evol : int
            PTO of evolution

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    cfs = import_pc_module(kind, esf.process)
    ihq = nl + 1
    m2hq = esf.info.m2hq[ihq - 4]
    # matching scale
    mu2hq = esf.info.threshold.area_walls[ihq - 3]
    # add normal terms starting from NNLO
    nnlo_terms = generate_heavy_diff(esf, nl, pto_evol)
    for k in nnlo_terms:
        k.min_order = 2
    if esf.process == "CC":
        w = kernels.cc_weights(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[ihq - 1], ihq
        )
        wq = {k: v for k, v in w["ns"].items() if abs(k) == ihq}
        if kind == "F3":
            return (
                -kernels.Kernel(
                    wq,
                    cfs.MatchingIntrinsicRplus(esf, nl, m1sq=m2hq, mu2hq=mu2hq),
                ),
                -kernels.Kernel(
                    {21: list(wq.values())[0]},
                    cfs.MatchingGluonRplus(esf, nl, m1sq=m2hq, mu2hq=mu2hq),
                ),
                *nnlo_terms,
            )
        return (
            -kernels.Kernel(
                wq, cfs.MatchingIntrinsicSplus(esf, nl, m1sq=m2hq, mu2hq=mu2hq)
            ),
            -kernels.Kernel(
                {21: list(wq.values())[0]},
                cfs.MatchingGluonSplus(esf, nl, m1sq=m2hq, mu2hq=mu2hq),
            ),
            *nnlo_terms,
        )
    # NC
    if kind == "F3":
        wVA = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "VA")
        wAV = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AV")
        wp = wVA + wAV
        wm = wVA - wAV
        return (
            -kernels.Kernel(
                {ihq: wp, (-ihq): -wp},
                cfs.MatchingIntrinsicRplus(esf, nl, m1sq=m2hq, m2sq=m2hq, mu2hq=mu2hq),
            ),
            -kernels.Kernel(
                {ihq: wm, (-ihq): -wm},
                cfs.MatchingIntrinsicRminus(esf, nl, m1sq=m2hq, m2sq=m2hq, mu2hq=mu2hq),
            ),
            *nnlo_terms,
        )
    # add matching terms
    wVV = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "VV")
    wAA = esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AA")
    wp = wVV + wAA
    wm = wVV - wAA
    return (
        -kernels.Kernel(
            {ihq: wp, (-ihq): wp},
            cfs.MatchingIntrinsicSplus(esf, nl, m1sq=m2hq, m2sq=m2hq, mu2hq=mu2hq),
        ),
        -kernels.Kernel(
            {ihq: wm, (-ihq): wm},
            cfs.MatchingIntrinsicSminus(esf, nl, m1sq=m2hq, m2sq=m2hq, mu2hq=mu2hq),
        ),
        # the explicit 2 is coming from Eq. (B.25) of :cite:`luca-intrinsic`.
        # it is coming from the sum over quark and anti-quark (a quark split in
        # both and either of them can interact with the EW boson)
        -kernels.Kernel(
            {21: 2 * wp},
            cfs.MatchingGluonSplus(esf, nl, m1sq=m2hq, m2sq=m2hq, mu2hq=mu2hq),
        ),
        -kernels.Kernel(
            {21: 2 * wm},
            cfs.MatchingGluonSminus(esf, nl, m1sq=m2hq, m2sq=m2hq, mu2hq=mu2hq),
        ),
        *nnlo_terms,
    )



# -*- coding: utf-8 -*-
import copy
import importlib
from numbers import Number

from eko import basis_rotation as br


def import_local(kind, process, sibling):
    """
    Import the suitable subpackage with the actual partonic channel implementation.

    Parameters
    ----------
        kind : str
            structure function kind
        process : str
            DIS process type: "EM","NC","CC"
        sibling : str
            relative parent to import from

    Returns
    -------
        module : module
            module
    """
    kind = kind.lower()
    process = process.lower()
    if process == "em":
        process = "nc"
    parent = ".".join(sibling.split(".")[:-1])
    return importlib.import_module(f".{kind}_{process}", parent)


class Kernel:
    """
    Combination of partons with their weights and their mathematical expression in this channel.

    Parameters
    ----------
        partons : dict
            mapping pid -> weight
        coeff : PartonicChannel
            mathematical expression
        max_order : int
            if given, silence above this order
        min_order : int
            if given, silenve below this order
    """

    def __init__(self, partons, coeff, max_order=None, min_order=None):
        self.partons = partons
        self.coeff = coeff
        self.max_order = max_order
        self.min_order = min_order

    def has_order(self, order):
        """
        Is current order active?

        Parameters
        ----------
            order : int
                order

        Returns
        -------
            bool :
                is active?
        """
        if self.min_order and order < self.min_order:
            return False
        if self.max_order and order > self.max_order:
            return False
        return True

    @property
    def channel(self):
        cls = str(type(self.coeff)).split("'")[1].split(".")[-1]

        # TODO: remove AsyQuark for AsyNonSinglet
        if "NonSinglet" in cls or "Quark" in cls:
            return "non-singlet"
        elif "Singlet" in cls:
            return "singlet"
        elif "Gluon" in cls:
            return "gluon"
        elif any([x in cls for x in ["Splus", "Sminus", "Rplus", "Rminus"]]):
            return "intrinsic"
        else:
            raise ValueError(f"Class '{cls}' does not correspond to a known channel")

    def __repr__(self):
        return repr({"partons": self.partons, "coeff": self.coeff})

    def __neg__(self):
        return self.__rmul__(-1)

    def __mul__(self, f):
        return self.__rmul__(f)

    def __rmul__(self, f):
        if not isinstance(f, Number):
            raise ValueError("Can only multiply numbers")
        partons = {k: f * v for k, v in self.partons.items()}
        return self.__class__(
            partons, copy.copy(self.coeff), self.max_order, self.min_order
        )


def cc_weights(coupling_constants, Q2, kind, cc_mask, nf):
    """
    Collect the weights of the partons.

    Parameters
    ----------
        coupling_constants : CouplingConstants
            manager for coupling constants
        Q2 : float
            W virtuality
        kind : str
            structure function kind
        cc_mask : str
            participating flavors on the CKM matrix
        nf : int
            number of light flavors

    Returns
    -------
        weights : dict
            mapping pid -> weight for q and g channel
    """
    weights = {"ns": {}, "g": {}, "s": {}}
    # determine couplings
    projectile_pid = coupling_constants.obs_config["projectilePID"]
    if projectile_pid in [-11, 12]:
        rest = 1
    else:
        rest = 0
    # quark couplings
    tot_ch_sq = 0
    norm = len(cc_mask)
    # iterate: include the heavy quark itself, since it can run in the singlet sector diagrams
    for q in range(1, min(nf + 2, 6 + 1)):
        sign = 1 if q % 2 == rest else -1
        w = coupling_constants.get_weight(q, Q2, None, cc_mask=cc_mask)
        # the heavy quark can not be in the input
        # NOTE: intrinsic abuse this statement with nf -> nf + 1
        if q <= nf:
            # @F3-sign@
            weights["ns"][sign * q] = w if kind != "F3" else sign * w
        # but it contributes to the average
        tot_ch_sq += w
    # gluon coupling = charge sum
    if rest == 0 and kind == "F3":
        tot_ch_sq *= -1
    weights["g"][21] = tot_ch_sq / norm / 2
    # add singlet
    for q in weights["ns"]:
        weights["s"][q] = tot_ch_sq / norm / 2
        weights["s"][-q] = tot_ch_sq / norm / 2
    return weights


def cc_weights_even(coupling_constants, Q2, kind, cc_mask, nf):
    """
    Collect the weights of the partons.

    Parameters
    ----------
        coupling_constants : CouplingConstants
            manager for coupling constants
        Q2 : float
            W virtuality
        kind : str
            structure function kind
        cc_mask : str
            participating flavors on the CKM matrix
        nf : int
            number of light flavors

    Returns
    -------
        weights : dict
            mapping pid -> weight for q and g channel
    """
    weights = {"ns": {}, "g": {}, "s": {}}
    # determine couplings
    projectile_pid = coupling_constants.obs_config["projectilePID"]
    if projectile_pid in [-11, 12]:
        rest = 1
    else:
        rest = 0
    # quark couplings
    tot_ch_sq = 0
    norm = len(cc_mask)
    # iterate: include the heavy quark itself, since it can run in the singlet sector diagrams
    for q in range(1, min(nf + 2, 6 + 1)):
        sign = 1 if q % 2 == rest else -1
        w = coupling_constants.get_weight(q, Q2, None, cc_mask=cc_mask)
        # the heavy quark can not be in the input
        # NOTE: intrinsic abuse this statement with nf -> nf + 1
        if q <= nf:
            # @F3-sign@
            weights["ns"][sign * q] = w / 2 * (1 if kind != "F3" else sign)
            weights["ns"][-sign * q] = w / 2 * (1 if kind != "F3" else sign)
        # but it contributes to the average
        tot_ch_sq += w
    # gluon coupling = charge sum
    weights["g"][21] = tot_ch_sq / norm / 2
    # add singlet
    for q in weights["ns"]:
        weights["s"][q] = tot_ch_sq / norm / 2
        weights["s"][-q] = tot_ch_sq / norm / 2
    return weights


def cc_weights_odd(coupling_constants, Q2, kind, cc_mask, nf):
    """
    Collect the weights of the partons.

    Parameters
    ----------
        coupling_constants : CouplingConstants
            manager for coupling constants
        Q2 : float
            W virtuality
        kind : str
            structure function kind
        cc_mask : str
            participating flavors on the CKM matrix
        nf : int
            number of light flavors

    Returns
    -------
        weights : dict
            mapping pid -> weight for q and g channel
    """
    weights = {"ns": {}}
    # determine couplings
    projectile_pid = coupling_constants.obs_config["projectilePID"]
    if projectile_pid in [-11, 12]:
        rest = 1
    else:
        rest = 0
    # quark couplings
    # iterate: include the heavy quark itself, since it can run in the singlet sector diagrams
    for q in range(1, min(nf + 2, 6 + 1)):
        sign = 1 if q % 2 == rest else -1
        w = coupling_constants.get_weight(q, Q2, None, cc_mask=cc_mask)
        # the heavy quark can not be in the input
        # NOTE: intrinsic abuse this statement with nf -> nf + 1
        if q <= nf:
            # @F3-sign@
            weights["ns"][sign * q] = w / 2 * (1 if kind != "F3" else sign)
            weights["ns"][-sign * q] = -w / 2 * (1 if kind != "F3" else sign)
    return weights


def generate_single_flavor_light(esf, nf, ihq):
    """
    Add a light-like contribution for a single quark flavor.

    The linear dependency to the electric charge is introduce by mulitplying
    and diving by nf. The multiplication is *implicit* inside the coefficient function,
    the division is *explict* made here.

    Parameters
    ----------
        esf : EvaluatedStructureFunction
            kinematic point
        nf : int
            number of light flavors
        ihq : int
            quark flavor to activate

    Returns
    -------
        elems : list(yadism.kernels.Kernel)
            list of elements
    """
    kind = esf.info.obs_name.kind
    light_cfs = import_local(
        kind, esf.process, ".".join(__name__.split(".")[:-1] + ["light", ""])
    )
    ns_partons = {}
    ch_av = 0
    s_partons = {}
    if esf.process == "CC":
        w_even = cc_weights_even(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[ihq - 1], nf
        )
        w_odd = cc_weights_odd(
            esf.info.coupling_constants, esf.Q2, kind, br.quark_names[ihq - 1], nf
        )
        return (
            Kernel(w_even["ns"], light_cfs.NonSingletEven(esf, nf)),
            Kernel({21: w_even["g"][21] / (nf)}, light_cfs.Gluon(esf, nf)),
            Kernel(
                {k: v / (nf) for k, v in w_even["s"].items()},
                light_cfs.Singlet(esf, nf),
            ),
            Kernel(w_odd["ns"], light_cfs.NonSingletOdd(esf, nf)),
        )
    if kind != "F3":
        w = esf.info.coupling_constants.get_weight(
            ihq, esf.Q2, "VV"
        ) + esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AA")
    else:
        w = esf.info.coupling_constants.get_weight(
            ihq, esf.Q2, "VA"
        ) + esf.info.coupling_constants.get_weight(ihq, esf.Q2, "AV")

    ns_partons[ihq] = w
    ns_partons[-ihq] = w if kind != "F3" else -w
    ch_av = w / (nf) if kind != "F3" else 0.0
    for pid in range(1, nf):
        s_partons[pid] = ch_av
        s_partons[-pid] = ch_av
    return (
        Kernel(ns_partons, light_cfs.NonSinglet(esf, nf)),
        Kernel({21: ch_av}, light_cfs.Gluon(esf, nf)),
        Kernel(s_partons, light_cfs.Singlet(esf, nf)),
    )



# -*- coding: utf-8 -*-
"""
Defines :py:class:`DistributionVec` and its API, that are used to represent
distribution objects in the coefficient function definition and calculation.

"""

import numpy as np
import scipy.integrate
from eko import interpolation

# TODO: don't use any default
eps_integration_border = 1e-10
"""Set the integration domain restriction, see
:ref:`Integration Note<integration-note>`"""

eps_integration_abs = 1e-13
"""Set the integration target absolute error, see
:ref:`Integration Note<integration-note>`"""


def quad_ker_reg_sing(z, x, is_log, areas, reg, reg_args, sing, pdf_at_x, sing_args):
    if is_log:
        pdf_at_x_ov_z_div_z = interpolation.log_evaluate_x(x / z, areas) / z
    else:
        pdf_at_x_ov_z_div_z = interpolation.evaluate_x(x / z, areas) / z
    # compute
    reg_integrand = reg(z, reg_args) * pdf_at_x_ov_z_div_z
    sing_integrand = sing(z, sing_args) * (pdf_at_x_ov_z_div_z - pdf_at_x)
    return reg_integrand + sing_integrand


def quad_ker_reg(z, x, is_log, areas, reg, reg_args):
    if is_log:
        pdf_at_x_ov_z_div_z = interpolation.log_evaluate_x(x / z, areas) / z
    else:
        pdf_at_x_ov_z_div_z = interpolation.evaluate_x(x / z, areas) / z
    # compute
    reg_integrand = reg(z, reg_args) * pdf_at_x_ov_z_div_z
    return reg_integrand


def quad_ker_sing(z, x, is_log, areas, sing, pdf_at_x, sing_args):
    if is_log:
        pdf_at_x_ov_z_div_z = interpolation.log_evaluate_x(x / z, areas) / z
    else:
        pdf_at_x_ov_z_div_z = interpolation.evaluate_x(x / z, areas) / z
    # compute
    sing_integrand = sing(z, sing_args) * (pdf_at_x_ov_z_div_z - pdf_at_x)
    return sing_integrand


def convolution(rsl, x, pdf_func):
    r"""
    Convolute a :py:class:`yadism.coefficient_functions.partonic_channel.RSL`
    instance with a function ``pdf_func``.

    The definition of the convolution performed is:

    .. math::
        \int_x^{1} \frac{\text{d}z}{z} dvec(z) f\left(\frac{x}{z}\right)

    (notice that is symmetryc in :math:`dvec \leftrightarrow f`).

    .. _integration-note:

    Note
    ----
    The module level attributes :py:attr:`eps_integration_abs` and
    :py:attr:`eps_integration_border` regulate the integration process,
    setting respectively the absolute error and restricting the
    integration domain in order to avoid singularities.

    Parameters
    ----------
    rsl: yadism.coefficient_functions.partonic_channel.RSL
        an object representing a distribution
    x : scalar
        the kinematics point at which the convoution is evaluated
    pdf_func : callable
        the function to be convoluted with ``rsl`` (usually a PDF, or a PDF
        interpolator)

    Returns
    -------
    float
        the result of the convolution
    float
        the integration error

    Note
    ----
    The real name of this function is ``convnd``.
    """
    # empty domain?
    if x >= (1 - eps_integration_border):
        return 0.0, 0.0
    # eko environment?
    if pdf_func.is_below_x(x):
        # support below x --> trivially 0
        return 0.0, 0.0
    else:
        # set breakpoints as relevant points of the integrand
        # computed from interpolation areas of `pdf_func`
        # and also eventually restrict integration domain
        area_borders = []
        for area in pdf_func.areas:
            area_borders.extend([area.xmin, area.xmax])
        area_borders = np.unique(area_borders)
        if pdf_func._mode_log:  # pylint: disable=protected-access
            area_borders = np.exp(area_borders)
        breakpoints = x / area_borders
        z_min = x
        z_max = min(max(breakpoints), 1)

    # cache values
    pdf_at_x = pdf_func(x)

    quad_ker = None
    quad_args = None
    if rsl.reg is not None or rsl.sing is not None:
        quad_args = (
            x,
            pdf_func._mode_log,
            pdf_func.areas_representation,
        )
        if rsl.reg is not None:
            reg_args = (
                rsl.reg,
                rsl.args["reg"],
            )
            quad_args = (*quad_args, *reg_args)
            quad_ker = quad_ker_reg
        if rsl.sing is not None:
            sing_args = (
                rsl.sing,
                pdf_at_x,
                rsl.args["sing"],
            )
            quad_args = (*quad_args, *sing_args)
            quad_ker = quad_ker_sing
        if rsl.reg is not None and rsl.sing is not None:
            quad_ker = quad_ker_reg_sing

    # actual convolution
    # ------------------

    # integrate the kernel, if needed
    if quad_ker is None:
        res, err = 0, 0
    else:
        res, err = scipy.integrate.quad(
            quad_ker,
            z_min * (1 + eps_integration_border),
            z_max * (1 - eps_integration_border),
            args=quad_args,
            epsabs=eps_integration_abs,
            points=breakpoints,
        )

    # sum the addends
    if rsl.loc is None:
        local_at_x = 0
    else:
        local_at_x = rsl.loc(x, rsl.args["loc"])

    res += pdf_at_x * local_at_x

    return res, err


def convolute_operator(fnc, interpolator):
    """
    Convolute function over all basis functions over all grid points.

    Parameters
    ----------
        fnc : RSL
            integration kernel
        interpolator : InterpolationDispatcher
            basis functions

    Returns
    -------
        ls : np.ndarray
            values
        els : np.ndarray
            errors
    """
    xgrid = interpolator.xgrid_raw
    grid_size = len(xgrid)
    op_res = np.zeros((grid_size, grid_size))
    op_err = np.zeros((grid_size, grid_size))
    # iterate output grid
    for k, xk in enumerate(interpolator.xgrid_raw):
        # iterate basis functions
        for l, bf in enumerate(interpolator):
            if k == l and l == grid_size - 1:
                continue
            # iterate sectors
            res, err = convolution(fnc, xk, bf)
            op_res[l, k] = res
            op_err[l, k] = err
    return op_res, op_err


def convolute_vector(cf, interpolator, convolution_point):
    """
    Convolute function over all basis functions.

    Parameters
    ----------
        cf : RSL
            integration kernel
        interpolator : InterpolationDispatcher
            basis functions
        convolution_point : float
            convolution point

    Returns
    -------
        ls : np.ndarray
            values
        els : np.ndarray
            errors
    """
    ls = []
    els = []
    # iterate all polynomials
    for polynomial_f in interpolator:
        c, e = convolution(cf, convolution_point, polynomial_f)
        ls.append(c)
        els.append(e)
    return np.array(ls), np.array(els)



# -*- coding: utf-8 -*-
from ..partonic_channel import RSL
from . import nlo, nnlo
from . import partonic_channel as pc


class NonSinglet(pc.LightBase):
    @staticmethod
    def LO():
        """
        |ref| implements :eqref:`4.2`, :cite:`vogt-f2nc`.
        """

        # leading order is just a delta function
        return RSL.from_delta(1.0)

    @staticmethod
    def NLO():
        """
        |ref| implements :eqref:`4.3`, :cite:`vogt-f2nc`.
        """

        return RSL.from_distr_coeffs(
            nlo.f2.ns_reg, (nlo.f2.ns_delta, nlo.f2.ns_omx, nlo.f2.ns_logomx)
        )

    def NNLO(self):
        """
        |ref| implements :eqref:`4.8`, :cite:`vogt-f2nc`.
        """

        return RSL(
            nnlo.xc2ns2p.c2nn2a, nnlo.xc2ns2p.c2ns2b, nnlo.xc2ns2p.c2nn2c, [self.nf]
        )


class Gluon(pc.LightBase):
    def NLO(self):
        r"""
        |ref| implements :eqref:`4.4`, :cite:`vogt-f2nc`.

        Note
        ----
        2 * n_f is coming from momentum sum
        rule q_i -> {q_i, g} but g -> {g, q_i, \bar{q_i} forall i}, so
        the 2 * n_f is needed to compensate for all the number of flavours
        plus antiflavours in which the gluon can go.
        """

        return RSL(nlo.f2.gluon_reg, args=[self.nf])

    def NNLO(self):
        """
        |ref| implements :eqref:`4.10`, :cite:`vogt-f2nc`.
        """

        return RSL(nnlo.xc2sg2p.c2g2a, loc=nnlo.xc2sg2p.c2g2c, args=[self.nf])


class Singlet(pc.LightBase):
    def NNLO(self):
        """
        |ref| implements :eqref:`4.9`, :cite:`vogt-f2nc`.
        """

        return RSL(nnlo.xc2sg2p.c2s2a, args=[self.nf])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'yadbox': 'src/yadbox',
 'yadbox.runcards': 'src/yadbox/runcards',
 'yadmark': 'src/yadmark',
 'yadmark.benchmark': 'src/yadmark/benchmark',
 'yadmark.benchmark.external': 'src/yadmark/benchmark/external',
 'yadmark.data': 'src/yadmark/data',
 'yadmark.navigator': 'src/yadmark/navigator'}

packages = \
['yadbox',
 'yadbox.runcards',
 'yadism',
 'yadism.coefficient_functions',
 'yadism.coefficient_functions.fonll',
 'yadism.coefficient_functions.heavy',
 'yadism.coefficient_functions.intrinsic',
 'yadism.coefficient_functions.light',
 'yadism.coefficient_functions.light.nlo',
 'yadism.coefficient_functions.light.nnlo',
 'yadism.coefficient_functions.special',
 'yadism.coefficient_functions.splitting_functions',
 'yadism.coefficient_functions.splitting_functions.nlo',
 'yadism.esf',
 'yadism.input',
 'yadmark',
 'yadmark.benchmark',
 'yadmark.benchmark.external',
 'yadmark.data',
 'yadmark.navigator']

package_data = \
{'': ['*']}

install_requires = \
['LeProHQ>=0.2.3,<0.3.0',
 'eko>=0.9.4,<0.10.0',
 'numba>=0.55.0,<0.56.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'rich>=12.4.4,<13.0.0',
 'scipy>=1.7.0,<2.0.0']

extras_require = \
{'docs': ['Sphinx>=4.1.1,<5.0.0',
          'sphinx-rtd-theme>=0.5.2,<0.6.0',
          'recommonmark>=0.7.1,<0.8.0',
          'sphinxcontrib-bibtex>=2.3.0,<3.0.0',
          'sphinxcontrib-details-directive>=0.1.0,<0.2.0',
          'nbsphinx>=0.8.6,<0.9.0'],
 'mark': ['banana-hep>=0.6.6,<0.7.0',
          'sqlalchemy>=1.4.21,<2.0.0',
          'a3b2bbc3ced97675ac3a71df45f55ba>=6.4.0,<7.0.0'],
 'pineappl': ['pineappl==0.5.0']}

entry_points = \
{'console_scripts': ['yadnav = yadmark.navigator:launch_navigator']}

setup_kwargs = {
    'name': 'yadism',
    'version': '0.11.4',
    'description': 'Yet Another Deep-Inelastic Scattering Module',
    'long_description': '<p align="center">\n  <a href="https://n3pdf.github.io/yadism/"><img alt="Yadism" src="https://raw.githubusercontent.com/N3PDF/yadism/master/docs/_assets/logo/logo.png" width=600></a>\n</p>\n\n<p align="center">\n  <a href="https://github.com/N3PDF/yadism/actions?query=workflow%3A%22yadism%22"><img alt="Tests" src="https://github.com/N3PDF/yadism/workflows/yadism/badge.svg"></a>\n  <a href=\'https://yadism.readthedocs.io/en/latest/?badge=latest\'><img src=\'https://readthedocs.org/projects/yadism/badge/?version=latest\' alt=\'Documentation Status\' /></a>\n  <a href="https://pypi.org/project/yadism/"><img alt="PyPI" src="https://img.shields.io/pypi/v/yadism"/></a>\n  <a href="https://codecov.io/gh/N3PDF/yadism"><img src="https://codecov.io/gh/N3PDF/yadism/branch/master/graph/badge.svg?token=qgCFyUQ6oG" /></a>\n  <a href="https://www.codefactor.io/repository/github/n3pdf/yadism"><img src="https://www.codefactor.io/repository/github/n3pdf/yadism/badge?s=e5a00668b58574b5b056e1aca01c7b25d2c203f8" alt="CodeFactor" /></a>\n  <a href="https://zenodo.org/badge/latestdoi/219968694"><img src="https://zenodo.org/badge/219968694.svg" alt="DOI"></a>\n</p>\n\n<!--Future Badges\n/github/workflow/status/N3PDF/dis/yadism\n\nuse the ones provided by shields.io:\n- example: https://img.shields.io/github/workflow/status/N3PDF/yadism\n\nnote: in order to make shields.io the repo must be public (or accessible to it in some way)\n\noptional:\n- Release-date (github):\n  - /github/release-date/:user/:repo\n- Downloads:\n  - github: /github/downloads/:user/:repo/total\n  - pypi: /pypi/:period/:packageName\n- License:\n  - pypi-license: /pypi/l/:packageName\n  - github-license: /github/license/:user/:repo\n- Activity:\n  - open-issues (github): /github/issues/:user/:repo\n  - open-pull-requests (github): /github/issues-pr/:user/:repo\n- Code size:\n  - github: /github/languages/code-size/:user/:repo\n-->\n\n## Scope of the project\n\nProvide all necessary tools to compute the DIS structure functions and related objects. This project is linked closely to [EKO](https://github.com/N3PDF/eko).\n\n## Installation\n\nAs a user please use [the released version on PyPI](https://pypi.org/project/yadism/),\nthrough your python package manager, e.g. with `pip`:\n\n```sh\npip install yadism\n```\n\n### Dev\n\nFor development just use [`poetry`](https://python-poetry.org/):\n\n```sh\npoetry install\n```\n\nTo install `poetry` and the other recommended tools, follow the\n[instructions](.github/CONTRIBUTING.md#development-tools).\n\n## Development\n\nMembers of the development team should always follow the [contribution\nguidelines](.github/contributing.md), to have a uniform strategy in code\ndevelopment and improve collaboration.\n\n## Contributing or contacting the authors\n\nFor any kind of interaction consider before to read [external contribution\nguidelines](.github/contributing.md#external-contributions), otherwise just send\nan email to the authors:\n\n- [Alessandro Candido](mailto:alessandro.candido@mi.infn.it)\n- [Felix Hekhorn](mailto:felix.hekhorn@mi.infn.it)\n- [Giacomo Magni](mailto:gmagni@nikhef.nl)\n',
    'author': 'Alessandro Candido',
    'author_email': 'alessandro.candido@mi.infn.it',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://n3pdf.github.io/yadism/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
from ..partonic_channel import RSL
from . import partonic_channel as pc


class Splus(pc.NeutralCurrentBase):
    def LO(self):
        factor = self.delta / self.ESF.Q2 * self.eta
        return RSL.from_delta(factor)

    def NLO(self):
        return self.mkNLO("2", "splus")


class Sminus(pc.NeutralCurrentBase):
    def NLO(self):
        return self.mkNLO("2", "sminus")



# -*- coding: utf-8 -*-
import numba as nb
import numpy as np
from eko.constants import CF, TR

zeta_2 = np.pi**2 / 6.0


@nb.njit("f8(f8,f8[:])", cache=True)
def ns_reg(z, args):
    # fmt: off
    return CF*(
        - 2 * (1 + z) * np.log((1 - z) / z)
        - 4 * np.log(z) / (1 - z)
        + 6 + 4 * z
    )
    # fmt: on


ns_delta = -CF * (9 + 4 * zeta_2)
ns_omx = -3 * CF
ns_logomx = 4 * CF


@nb.njit("f8(f8,f8[:])", cache=True)
def gluon_reg(z, args):
    nf = args[0]
    return (
        nf
        * (
            (2.0 - 4.0 * z * (1.0 - z)) * np.log((1.0 - z) / z)
            - 2.0
            + 16.0 * z * (1.0 - z)
        )
        * (2.0 * TR)
    )



# -*- coding: utf-8 -*-
"""
.. todo::
    docs
"""

from . import output, version
from .runner import Runner

__version__ = version.__version__


def run_yadism(theory: dict, observables: dict):
    """
    .. todo::
        - decide the purpose
        - implement
        - docs
    """
    runner = Runner(theory, observables)
    return runner.get_result()



# -*- coding: utf-8 -*-
import numba as nb
import numpy as np
from eko import constants

from .. import partonic_channel as pc
from .. import splitting_functions as split
from ..partonic_channel import RSL
from ..special import li2
from ..special.nielsen import nielsen
from ..splitting_functions import lo


class PartonicChannelAsy(pc.PartonicChannel):
    def __init__(self, *args, mu2hq):
        super().__init__(*args)
        self.L = np.log(self.ESF.Q2 / mu2hq)


# we can define those here, since F2=F3=delta(1-z) at LO and FL=0


@nb.njit("f8(f8)", cache=True)
def Delta_qq_sing(z):
    r"""
    |ref| implements :eqref:`101`, :cite:`forte-fonll`.

    Parameters
    ----------
        z : float
            parton momentum

    Returns
    -------
        singular part of : math:`\Delta_qq(z)`
    """
    as_norm = 4.0
    return (
        constants.CF
        * constants.TR
        * (
            (1.0 + z**2) / (1.0 - z) * (2.0 / 3.0 * np.log(z) + 10.0 / 9.0)
            + 4.0 / 3.0 * (1.0 - z)
        )
        * as_norm
    )


@nb.njit("f8(f8)", cache=True)
def Delta_qq_loc(x):
    r"""
    |ref| implements :eqref:`101`, :cite:`forte-fonll`.

    Parameters
    ----------
        x : float
            Bjorken x

    Returns
    -------
        local part of : math:`\Delta_qq(z)`
    """
    as_norm = 4.0
    # Integrate[(1+z^2)/(1-z)(2/3 Log[z]+10/9)+4/3(1-z),{z,0,x},Assumptions->{0<x<1}]
    return (
        constants.CF
        * constants.TR
        * (
            -4 * np.pi**2
            + (16 - 19 * x) * x
            - 40 * np.log(1 - x)
            - 6 * x * (2 + x) * np.log(x)
            + 24 * li2(1 - x)
        )
        / 18.0
        * as_norm
    )


@nb.njit("f8(f8, f8[:])", cache=True)
def K_qq_sing(z, _args):
    """
    |ref| implements :eqref:`100`, :cite:`forte-fonll`.

    Parameters
    ----------
        z : float
            parton momentum

    Returns
    -------
        singular part of : math:`K_qq(z)`
    """
    as_norm = 4.0
    return (
        constants.CF
        * constants.TR
        * (
            (1.0 + z**2)
            / (1.0 - z)
            * (1.0 / 6.0 * np.log(z) ** 2 + 5.0 / 9.0 * np.log(z) + 28.0 / 27.0)
            + (1.0 - z) * (2.0 / 3.0 * np.log(z) + 13.0 / 9.0)
        )
        * as_norm
    )


@nb.njit("f8(f8,f8[:])", cache=True)
def K_qq_loc(x, _args):
    """
    |ref| implements :eqref:`100`, :cite:`forte-fonll`.

    Parameters
    ----------
        x : float
            Bjorken x

    Returns
    -------
        local part of : math:`K_qq(z)`
    """
    as_norm = 4.0
    # Integrate[(1+z^2)/(1-z)(1/6 Log[z]^2+5/9Log[z]+28/27)+(1-z)(2/3Log[z]+13/9),{z,0,x},Assumptions->{0<x<1}]
    return (
        constants.CF
        * constants.TR
        * (
            -40 * np.pi**2
            - x * (8 + 211 * x)
            - 6
            * np.log(x)
            * (4 * np.pi**2 + x * (-16 + 19 * x) + 3 * x * (2 + x) * np.log(x))
            + 8 * np.log(1 - x) * (-56 + 9 * np.log(x) ** 2)
            + 48 * (5 + 3 * np.log(x)) * li2(1 - x)
            + 144 * nielsen(2, 1, x).real
        )
        / 216.0
        * as_norm
    )


@nb.njit("f8(f8,f8[:])", cache=True)
def pdf_matching_LL_reg(z, args):
    L = args[0]
    as_norm = 2.0
    return L**2 / 2.0 * 2.0 * constants.TR / 3 * as_norm * lo.pqq_reg(z, args)


@nb.njit("f8(f8,f8[:])", cache=True)
def pdf_matching_LL_sing(z, args):
    L = args[0]
    as_norm = 2.0
    return +(L**2) / 2.0 * 2.0 * constants.TR / 3 * as_norm * lo.pqq_sing(z, args)


@nb.njit("f8(f8,f8[:])", cache=True)
def pdf_matching_LL_loc(z, args):
    L = args[0]
    as_norm = 2.0
    return +(L**2) / 2.0 * 2.0 * constants.TR / 3 * as_norm * lo.pqq_local(z, args)


@nb.njit("f8(f8,f8[:])", cache=True)
def pdf_matching_NLL_sing(z, args):
    L = args[0]
    return -L * Delta_qq_sing(z)


@nb.njit("f8(f8,f8[:])", cache=True)
def pdf_matching_NLL_loc(z, args):
    L = args[0]
    return -L * Delta_qq_loc(z)


class PdfMatchingLLNonSinglet(PartonicChannelAsy):
    def NNLO(self):
        return RSL(
            pdf_matching_LL_reg,
            pdf_matching_LL_sing,
            pdf_matching_LL_loc,
            args=[self.L],
        )


class PdfMatchingNLLNonSinglet(PartonicChannelAsy):
    def NNLO(self):
        return RSL(sing=pdf_matching_NLL_sing, loc=pdf_matching_NLL_loc, args=[self.L])


class PdfMatchingNNLLNonSinglet(PartonicChannelAsy):
    def NNLO(self):
        return RSL(sing=K_qq_sing, loc=K_qq_loc, args=[self.L])


class PartonicChannelAsyIntrinsic(pc.PartonicChannel):
    def __init__(self, *args, m1sq, m2sq):
        super().__init__(*args)
        self.Q2 = self.ESF.Q2
        self.x = self.ESF.x
        self.m1sq = m1sq
        self.m2sq = m2sq
        self.sigma_pm = self.Q2 + self.m2sq - self.m1sq
        self.delta = self.kinematic_delta(self.m1sq, self.m2sq, -self.Q2)
        self.eta = 2.0 * self.Q2 / (self.sigma_pm + self.delta)

    @staticmethod
    def kinematic_delta(a, b, c):
        return np.sqrt(a**2 + b**2 + c**2 - 2 * (a * b + b * c + c * a))

    def convolution_point(self):
        return self.x / self.eta


class FMatching(PartonicChannelAsyIntrinsic):
    ffns = lambda *_args, m1sq, m2sq: None

    def __init__(self, *args, m1sq, m2sq, mu2hq):
        super().__init__(*args, m1sq=m1sq, m2sq=m2sq)
        self.l = np.log(self.Q2 / mu2hq)

    def obj(self):
        return self.ffns(self.ESF, self.nf, m1sq=self.m1sq, m2sq=self.m2sq)

    def parent_lo_local(self):
        parent_LO = self.obj().LO()
        if parent_LO is None:
            return None
        return parent_LO.args["loc"][0]


class FMatchingQuark(FMatching):
    def NLO(self):
        icl = self.parent_lo_local()
        if icl is None:
            return None
        asnorm = 2.0
        l = self.l

        def sing(z, _args):
            # this coefficient function is *almost* proportional to P_qq
            # i.e. 2CF * (1.0 + z ** 2) / (1.0 - z) is the "bare" P_qq
            return (
                asnorm
                * icl
                * constants.CF
                * ((1.0 + z**2) / (1.0 - z) * (l - 2.0 * np.log(1.0 - z) - 1.0))
            )

        # MMa:
        # FortranForm@FullSimplify@Integrate[(1 + z^2)/(1 - z) (l - 2 Log[1 - z] - 1), {z, 0, x}, Assumptions -> {0 < x < 1}] # pylint: disable=line-too-long
        def loc(x, _args):
            return (
                -asnorm
                * icl
                * constants.CF
                * (
                    -(x * (4.0 + l * (2.0 + x))) / 2.0
                    + np.log(1.0 - x)
                    * (-1.0 - 2.0 * l + x * (2.0 + x) + 2.0 * np.log(1.0 - x))
                )
            )

        return RSL(sing=sing, loc=loc)


class FMatchingCC(FMatching):
    ffns = lambda *_args, m1sq: None

    def __init__(self, *args, m1sq, mu2hq):
        super().__init__(*args, m1sq=m1sq, m2sq=0.0, mu2hq=mu2hq)

    def obj(self):
        return self.ffns(self.ESF, self.nf, m1sq=self.m1sq)


class FMatchingQuarkCC(FMatchingCC, FMatchingQuark):
    pass


class FMatchingGluon(FMatching):
    def NLO(self):
        return self.mk_nlo_raw(self.l)

    def mk_nlo_raw(self, l):
        icl = self.parent_lo_local()
        if icl is None:
            return None

        # since as and p_qg appear together there is no need to put an explicit as_norm here
        def reg(z, _args):
            return icl * split.lo.pqg_single(z, np.array([], dtype=float)) * l

        return RSL(reg)


class FMatchingGluonCC(FMatchingCC, FMatchingGluon):
    pass
